﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Microsoft.Xrm.Sdk;
using Microsoft.Xrm.Sdk.Query;
using Microsoft.Xrm.Sdk.Messages;
using Microsoft.Crm.Sdk.Messages;
using mcsScheduling;
using MCSUtilities2011;
using System.ServiceModel;

namespace MCSShared
{
    public static class CvtHelper
    {
        #region Helper Functions

        /// <summary>
        /// This accepts 1 or 2 strings and outputs them as a clickable link
        /// </summary>
        /// <param name="input">the string to be converted to a URL</param>
        /// <param name="clickDisplay">the display text that is clickable - defaults to be the same as the actual URL unless otherwise specified</param>
        /// <returns>formatted url</returns>
        public static string buildHTMLUrl(string input, string clickDisplay = "")
        {
            var http = (input.Length < 4 || input.Substring(0, 4).ToLower() != "http") ? "https://" + input : input;
            clickDisplay = (clickDisplay == string.Empty) ? http : clickDisplay;
            return "<a href=" + http + ">" + clickDisplay + "</a>";
        }
        /// <summary>
        /// Create a note related to this object
        /// </summary>
        /// <param name="Regarding"></param>
        /// <param name="Message"></param>
        /// <param name="Title"></param>
        /// <param name="OrgService"></param>
        public static void CreateNote(EntityReference Regarding, String Message, String Title, IOrganizationService OrgService)
        {
            var newNote = new Entity("annotation");
            newNote["subject"] = Title;
            newNote["notetext"] = Message;
            newNote["objectid"] = Regarding;
            OrgService.Create(newNote);
        }

        /// <summary>
        /// Single Sign On URL conversion
        /// </summary>
        /// <param name="PluginExecutionContext"></param>
        /// <returns></returns>
        public static string getServerURL(IPluginExecutionContext PluginExecutionContext)
        {
            string url = "https://";
            string orgname = PluginExecutionContext.OrganizationName;
            switch (orgname.ToUpper())
            {
                case "CVT":
                    //PROD
                    url += "internalsrv.xrm.domain/CVT";
                    break;
                case "QACVT":
                    //QA
                    url += "internalsrv.np.crm.vrm.vba.domain/QACVT";
                    break;
                case "INTCVT":
                    //INT
                    url += "internalsrv.np.crm.vrm.vba.domain/INTCVT";
                    break;
                case "PRECVT":
                    //Pre-Prod
                    url += "internalsrv.np.crm.vrm.vba.domain/PRECVT";
                    break;
                default:
                    //PROD
                    url += "internalsrv.xrm.domain/CVT";
                    break;
            }
            return url;
        }

        /// <summary>
        /// Retrieves the ETC
        /// </summary>
        /// <param name="service"></param>
        /// <param name="entity"></param>
        /// <returns></returns>
        public static int GetEntityTypeCode(IOrganizationService service, Entity entity)
        {
            return GetEntityTypeCode(service, entity.LogicalName);
        }

        /// <summary>
        /// Get Entity Type Code because it can change per environment
        /// </summary>
        /// <param name="service"></param>
        /// <param name="entityname"></param>
        /// <returns></returns>
        public static int GetEntityTypeCode(IOrganizationService service, String entityname)
        {
            RetrieveEntityRequest request = new RetrieveEntityRequest();
            request.LogicalName = entityname;
            RetrieveEntityResponse response = (RetrieveEntityResponse)service.Execute(request);
            int objectTypecode = response.EntityMetadata.ObjectTypeCode.Value;
            return objectTypecode;
        }

        /// <summary>
        /// This method accepts the resource (or user) id from the patient or provider site resource, and then returns the user or equipment record
        /// </summary>
        /// <param name="id"></param>
        /// <param name="context"></param>
        /// <returns></returns>
        public static Entity getPartyParticipantFromResource(Guid id, Xrm context)
        {
            var participant = new Entity();
            participant = context.EquipmentSet.FirstOrDefault(e => e.mcs_relatedresource.Id == id);
            if (participant == null)
                participant = context.SystemUserSet.FirstOrDefault(u => u.Id == id);
            return participant;
        }

        /// <summary>
        /// Get owner of workflow
        /// </summary>
        /// <param name="WorkflowName"></param>
        /// <param name="OrganizationService"></param>
        /// <returns></returns>
        public static List<ActivityParty> GetWorkflowOwner(string WorkflowName, IOrganizationService OrganizationService)
        {
            //Get the workflow definition's owner
            using (var context = new Xrm(OrganizationService))
            {
                var workflow = ((from w in context.WorkflowSet
                        where (w.Type.Value == 1 &&
                        w.Name == WorkflowName)
                        select new
                        {
                            w.OwnerId
                        }).FirstOrDefault());

                //var workflow = context.WorkflowSet.FirstOrDefault(c =>
                //    c.Type.Value == 1 && c.Name == WorkflowName);
                
                if (workflow != null && workflow.OwnerId != null)
                {
                    var owner = new ActivityParty()
                    {
                        PartyId = workflow.OwnerId
                    };
                    return CvtHelper.SetPartyList(owner);
                }
                else
                {
                    return null;
                }
            }
        }
        #endregion

        #region Email Related Functions
        //Overloaded method: creates a list from an Activty Party to populate a From, To, or CC field
        public static List<ActivityParty> SetPartyList(ActivityParty user)
        {
            List<ActivityParty> userList = new List<ActivityParty>();
            //user if exists
            if (user != null)
                userList.Add(user);
            return userList;
        }

        //Overloaded method: return empty list if userRef is null, otherwise get Activity party and call first SetEmailProperties method
        public static List<ActivityParty> SetPartyList(EntityReference userRef)
        {
            return userRef == null ? new List<ActivityParty>() :
                SetPartyList(new ActivityParty() { PartyId = new EntityReference(userRef.LogicalName, userRef.Id) });
        }

        //Overloaded method: creates a list from an EntityCollection to populate a From, To, or CC field
        public static List<ActivityParty> SetPartyList(EntityCollection userCollection)
        {
            List<ActivityParty> userList = new List<ActivityParty>();

            foreach (var record in userCollection.Entities.Select(e => e).Distinct())
            {
                var activityParty = new ActivityParty() { PartyId = new EntityReference(record.LogicalName, record.Id) };
                userList.Add(activityParty);
            }
            return userList;
        }

        //This method takes the email passed in and sends it
        public static void SendEmail(Email email, IOrganizationService OrganizationService)
        {
            SendEmailRequest requestObj = new SendEmailRequest()
            {
                EmailId = (Guid)email.ActivityId,
                IssueSend = true,
                TrackingToken = ""
            };
            SendEmailResponse response = (SendEmailResponse)OrganizationService.Execute(requestObj);
        }

        //Saves and Sends the email
        public static void UpdateSendEmail(Email email, IOrganizationService OrganizationService)
        {
            OrganizationService.Update(email);
            SendEmail(email, OrganizationService);
        }
        #endregion

        #region MTSA/TSA/Service Related Functions
        /// <summary>
        /// Calls CreateUpdateService with TSA guid
        /// </summary>
        /// <param name="tsaGuid"></param>
        /// <param name="Logger"></param>
        /// <param name="OrganizationService"></param>
        /// <param name="McsSettings"></param>
        internal static void UpdateTSAFields(Guid tsaGuid, MCSLogger Logger, IOrganizationService OrganizationService, MCSSettings McsSettings)
        {
            Logger.WriteDebugMessage("Verifying TSA to update");
            try
            {
                if (tsaGuid != null)
                {
                    Logger.WriteDebugMessage("Updating the TSA's Service components to reflect changes.");
                    CvtHelper.CreateUpdateService(tsaGuid, Logger, OrganizationService, McsSettings);
                    Logger.WriteDebugMessage("TSA Updated");
                }
            }
            catch (Exception ex)
            {
                Logger.WriteToFile(ex.Message);
                throw new InvalidPluginExecutionException(ex.Message);
            }
        }

        /// <summary>
        /// Calls UpdateMTSA with MTSA guid
        /// </summary>
        /// <param name="mtsaGuid"></param>
        /// <param name="Logger"></param>
        /// <param name="OrganizationService"></param>
        internal static void UpdateMTSAFields(Guid mtsaGuid, MCSLogger Logger, IOrganizationService OrganizationService)
        {
            Logger.WriteDebugMessage("Verifying MTSA to update");
            try
            {
                if (mtsaGuid != null)
                {
                    Logger.WriteDebugMessage("Updating the MTSA");
                    CvtHelper.UpdateMTSA(mtsaGuid, Guid.Empty, Logger, OrganizationService);
                    Logger.WriteDebugMessage("MTSA Updated");
                }
            }
            catch (Exception ex)
            {
                Logger.WriteToFile(ex.Message);
                throw new InvalidPluginExecutionException(ex.Message);
            }
        }

        /// <summary>
        /// Update TSAs provider and vista clinic string fields and build the service if it is in status of Production.
        /// </summary>
        /// <param name="TSAId"></param>
        /// <param name="Logger"></param>
        /// <param name="OrganizationService"></param>
        /// <param name="McsSettings"></param>
        /// <remarks>3-25-2015: Major Refactoring to improve readability and consolidate code.  Errors currently exist, so this should fix them as well</remarks>
        public static void CreateUpdateService(Guid TSAId, MCSLogger Logger, IOrganizationService OrganizationService, MCSSettings McsSettings)
        {
            Logger.setMethod = "CvtHelper.CreateUpdateService";
            Logger.WriteDebugMessage("starting CreateUpdateService");

            using (var srv = new Xrm(OrganizationService))
            {
                try
                {
                    if (TSAId == Guid.Empty)
                    {
                        Logger.WriteDebugMessage("No TSA value, exiting.");
                        return;
                    }

                    //Get current TSA
                    var thisTSA = srv.mcs_servicesSet.FirstOrDefault(i => i.Id == TSAId);

                    Logger.WriteDebugMessage("Starting to build the Service components");

                    //Declaration of Variables
                    StringBuilder builder = new System.Text.StringBuilder("");
                    var builderProv = new System.Text.StringBuilder("");
                    var builderProvAllRequired = new System.Text.StringBuilder("");
                    var builderPat = new System.Text.StringBuilder("");
                    var builderPatAllRequired = new System.Text.StringBuilder("");
                    string provSiteVCs = null;
                    string providers = null;
                    string patSiteVCs = null;
                    string patSites = null;
                    var tsaType = "Clinic";
                    Guid resourceSpecId;

                    #region PROVIDER SITE RESOURCES
                    //Retrieve all the provider site resources
                    Logger.WriteDebugMessage("Start Sorting through Provider Site Resources");
                    var getProvResources = from provGroups in srv.cvt_providerresourcegroupSet
                                           where provGroups.cvt_RelatedTSAid.Id == TSAId
                                           where provGroups.statecode == 0
                                           select new
                                           {
                                               provGroups.Id,
                                               provGroups.cvt_RelatedResourceId,
                                               provGroups.cvt_resourcespecguid,
                                               provGroups.cvt_TSAResourceType,
                                               provGroups.cvt_Type,
                                               provGroups.cvt_RelatedUserId,
                                               provGroups.cvt_RelatedResourceGroupid
                                           };

                    //Loop through all of the Provider Site resources
                    foreach (var provGroups in getProvResources)
                    {
                        //Verify that the Resource is typed, not required, but should be filled in
                        if (provGroups.cvt_Type != null)
                        {
                            //StrBuilder
                            if (provGroups.cvt_resourcespecguid != null)
                            {
                                switch (provGroups.cvt_Type.Value)
                                {
                                    case 917290000: //All Required
                                        builderProvAllRequired.Append(AddResourceToConstraintGroup(provGroups.cvt_resourcespecguid));

                                        break;
                                    default: //Others
                                        builderProv.Append(AddResourceToConstraintGroup(provGroups.cvt_resourcespecguid));
                                        break;
                                }
                            }
                            //Naming
                            switch (provGroups.cvt_Type.Value)
                            {
                                case 251920000: //Vista Clinic
                                    if (provGroups.cvt_TSAResourceType.Value == 0) //Group of Vista Clinics
                                    {
                                        //Query for child names
                                        var groupResourceRecords = srv.mcs_groupresourceSet.Where(g => g.mcs_relatedResourceGroupId.Id == provGroups.cvt_RelatedResourceGroupid.Id && g.statecode == 0);
                                        foreach (var child in groupResourceRecords)
                                        {
                                            if (child.mcs_RelatedResourceId != null)
                                                provSiteVCs += child.mcs_RelatedResourceId.Name.ToString() + " ; ";
                                        }
                                    }
                                    else
                                    {
                                        if (provGroups.cvt_RelatedResourceId != null)
                                            provSiteVCs += provGroups.cvt_RelatedResourceId.Name.ToString() + " ; ";
                                    }
                                    break;
                                case 100000000: //Telepresenter/Imager
                                case 99999999: //Provider
                                    if (provGroups.cvt_TSAResourceType.Value == 0) //Group of Providers
                                    {
                                        //Query for child names
                                        var groupResourceRecords = srv.mcs_groupresourceSet.Where(g => g.mcs_relatedResourceGroupId.Id == provGroups.cvt_RelatedResourceGroupid.Id && g.statecode == 0);
                                        foreach (var child in groupResourceRecords)
                                        {
                                            if (child.mcs_RelatedUserId != null)
                                                providers += child.mcs_RelatedUserId.Name.ToString() + " ; ";
                                        }
                                    }
                                    else
                                    {
                                        if (provGroups.cvt_RelatedUserId != null)
                                            providers += provGroups.cvt_RelatedUserId.Name.ToString() + " ; ";
                                    }
                                    break;
                                case 917290000: //All Required
                                    //Query for child names
                                    var childgroupResourceRecords = srv.mcs_groupresourceSet.Where(g => g.mcs_relatedResourceGroupId.Id == provGroups.cvt_RelatedResourceGroupid.Id && g.statecode == 0);
                                    foreach (var child in childgroupResourceRecords)
                                    {
                                        if (child.mcs_RelatedUserId != null)
                                            providers += child.mcs_RelatedUserId.Name.ToString() + " ; ";
                                        else
                                        { //Check the related Resource if it is a Vista Clinic
                                            var childR = srv.mcs_resourceSet.First(r => r.Id == child.mcs_RelatedResourceId.Id);
                                            if (childR.mcs_Type.Value == 251920000)
                                                provSiteVCs += childR.mcs_name + " ; ";
                                        }
                                    }
                                    break;
                            }
                        }
                        else //Probably Single Provider, but check.
                        {
                            //Provider or Telepresenter
                            if ((provGroups.cvt_TSAResourceType.Value == 2) || (provGroups.cvt_TSAResourceType.Value == 3))
                            {
                                if (provGroups.cvt_RelatedUserId != null)
                                {
                                    builderProv.Append(AddResourceToConstraintGroup(provGroups.cvt_resourcespecguid));
                                    providers += provGroups.cvt_RelatedUserId.Name.ToString() + " ; ";
                                }
                            }
                        }
                    }
                    Logger.WriteDebugMessage("Finished Sorting through Provider Site Resources");
                    #endregion

                    #region PATIENT SITE RESOURCES
                    //Retrieve all the patient site resources
                    Logger.WriteDebugMessage("Start Sorting through Patient Site Resources");
                    var getPatResources = from patGroups in srv.cvt_patientresourcegroupSet
                                          where patGroups.cvt_RelatedTSAid.Id == TSAId
                                          where patGroups.statecode == 0
                                          select new
                                          {
                                              patGroups.Id,
                                              patGroups.cvt_RelatedResourceId,
                                              patGroups.cvt_resourcespecguid,
                                              patGroups.cvt_TSAResourceType,
                                              patGroups.cvt_type,
                                              patGroups.cvt_RelatedResourceGroupid
                                          };

                    //Loop through all of the Patient Site resources
                    foreach (var patGroups in getPatResources)
                    {
                        //Verify that the Resource is typed
                        if (patGroups.cvt_type != null)
                        {
                            //Check that ResourceSpecGuid is not null
                            if (patGroups.cvt_resourcespecguid != null)
                            {
                                switch (patGroups.cvt_type.Value)
                                {
                                    case 917290000: //All Required
                                        builderPatAllRequired.Append(AddResourceToConstraintGroup(patGroups.cvt_resourcespecguid));
                                        //Query for child names, if Vista Clinics then write into the string
                                        var childgroupResourceRecords = srv.mcs_groupresourceSet.Where(g => g.mcs_relatedResourceGroupId.Id == patGroups.cvt_RelatedResourceGroupid.Id && g.statecode == 0);
                                        foreach (var child in childgroupResourceRecords)
                                        {
                                            if (child.mcs_RelatedUserId == null)
                                            {
                                                //Check the related Resource if it is a Vista Clinic
                                                var childR = srv.mcs_resourceSet.First(r => r.Id == child.mcs_RelatedResourceId.Id);
                                                if (childR.mcs_Type.Value == 251920000)
                                                    patSiteVCs += childR.mcs_name + " ; ";
                                            }
                                        }
                                        break;
                                    case 251920000: //Vista Clinic
                                        builderPat.Append(AddResourceToConstraintGroup(patGroups.cvt_resourcespecguid));
                                        if (patGroups.cvt_RelatedResourceId != null)
                                            patSiteVCs += patGroups.cvt_RelatedResourceId.Name.ToString() + " ; ";
                                        break;
                                    default: //Others                           
                                        builderPat.Append(AddResourceToConstraintGroup(patGroups.cvt_resourcespecguid));
                                        break;
                                }
                            }
                        }
                        //If not typed, could be single provider, shouldn't be on patient side, but if
                        else
                        {   //Provider or Telepresenter
                            if ((patGroups.cvt_TSAResourceType.Value == 2) || (patGroups.cvt_TSAResourceType.Value == 3))
                            {
                                if (patGroups.cvt_resourcespecguid != null)
                                {
                                    builderPat.Append(AddResourceToConstraintGroup(patGroups.cvt_resourcespecguid));
                                }
                            }
                        }
                    }

                    //Getting all the unique Patient Sites to add into a field for the view.                      
                    var getPatientSites = (from patGroupSites in srv.cvt_patientresourcegroupSet
                                           where patGroupSites.cvt_RelatedTSAid.Id == TSAId
                                           where patGroupSites.statecode == 0
                                           select new
                                           {
                                               patGroupSites.cvt_relatedsiteid,
                                           }).Distinct();

                    foreach (var patGroupSites in getPatientSites)
                    {
                        if (patGroupSites.cvt_relatedsiteid != null)
                            patSites += patGroupSites.cvt_relatedsiteid.Name.ToString() + " ; ";
                    }
                    Logger.WriteDebugMessage("Finished Sorting through Patient Site Resources");
                    #endregion

                    //Check for TSA status != Production
                    if (thisTSA.statuscode.Value != 251920000)
                    {
                        Logger.WriteDebugMessage("TSA is not in status Production, just update the string fields");
                        UpdateTSA(thisTSA, patSites, providers, patSiteVCs, provSiteVCs, Logger, OrganizationService, srv, Guid.Empty);
                        return;
                    }

                    //Else continue building
                    #region Logic - Constructing the builders etc
                    //Validation: No Resources, throw error
                    if (builderProv.Length == 0 && builderProvAllRequired.Length == 0 && builderPat.Length == 0 && builderPatAllRequired.Length == 0)
                        throw new InvalidPluginExecutionException("A TSA must have a resource listed in order to be put into production");
                    //Determine the Type of TSA, and check specifically
                    if (thisTSA.cvt_Type == true)
                        tsaType = "Home";
                    if (thisTSA.cvt_AvailableTelehealthModalities != null && thisTSA.cvt_AvailableTelehealthModalities.Value == 917290001)
                        tsaType = "SFT";
                    if (thisTSA.cvt_groupappointment == true)
                        tsaType = "Group";

                    switch (tsaType)
                    {
                        case "Home"://Home/Mobile - only Provider Site
                            if (builderProv.Length == 0 && builderProvAllRequired.Length == 0)
                                throw new InvalidPluginExecutionException("A Home/Mobile TSA must have provider resources Listed in order to be put into production");
                            else
                            {
                                if (builderProv.Length > 0)
                                    builder.Append(builderProv);

                                if (builderProvAllRequired.Length > 0)
                                {
                                    Guid ProvAllRequiredSpecId = BuildOut(thisTSA, builderProvAllRequired, OrganizationService, 1, 0, true);
                                    builder.Append(AddResourceToConstraintGroup(ProvAllRequiredSpecId.ToString()));
                                }
                            }
                            break;
                        case "SFT"://SFT - only Patient Site
                            if (builderPat.Length == 0 && builderPatAllRequired.Length == 0)
                                throw new InvalidPluginExecutionException("A Store Forward TSA must have patient resources listed in order to be put into production");
                            else
                            {
                                if (builderPat.Length > 0)
                                    builder.Append(builderPat);

                                if (builderPatAllRequired.Length > 0)
                                {
                                    Guid PatAllRequiredSpecId = BuildOut(thisTSA, builderPatAllRequired, OrganizationService, 1, 0, true);
                                    builder.Append(AddResourceToConstraintGroup(PatAllRequiredSpecId.ToString()));
                                }
                            }
                            break;
                        default: // Clinic Based or Group (PatAllReq is all) - both Provider and Patient Site
                            if (builderProv.Length == 0 && builderProvAllRequired.Length == 0)
                                throw new InvalidPluginExecutionException("A TSA must have a provider resource listed in order to be put into production");
                            if (builderPat.Length == 0 && builderPatAllRequired.Length == 0)
                                throw new InvalidPluginExecutionException("A TSA must have a patient resource listed in order to be put into production");
                            if (builderProv.Length > 0)
                                builder.Append(builderProv);

                            if (builderProvAllRequired.Length > 0)
                            {
                                Guid ProvAllRequiredSpecId = BuildOut(thisTSA, builderProvAllRequired, OrganizationService, 1, 0, true);
                                builder.Append(AddResourceToConstraintGroup(ProvAllRequiredSpecId.ToString()));
                            }

                            if (builderPat.Length > 0)
                                builder.Append(builderPat);

                            if (builderPatAllRequired.Length > 0)
                            {
                                Guid PatAllRequiredSpecId;
                                if (tsaType == "Group")
                                    PatAllRequiredSpecId = BuildOut(thisTSA, builderPatAllRequired, OrganizationService, -1, 0, true);
                                else
                                    PatAllRequiredSpecId = BuildOut(thisTSA, builderPatAllRequired, OrganizationService, 1, 0, true);

                                builder.Append(AddResourceToConstraintGroup(PatAllRequiredSpecId.ToString()));
                            }
                            break;
                    }
                    #endregion

                    #region Logic - Constructing the service
                    Logger.WriteDebugMessage("Building out ResourceSpec");
                    resourceSpecId = BuildOut(thisTSA, builder, OrganizationService, -1, 2, false);

                    Service newService = new Service
                    {
                        Name = thisTSA.mcs_name.ToString(),
                        AnchorOffset = 480,
                        Duration = thisTSA.cvt_duration.Value,
                        InitialStatusCode = new OptionSetValue(4),
                        Granularity = "FREQ=MINUTELY;INTERVAL=" + thisTSA.cvt_startevery.ToString() + ";",
                        ResourceSpecId = new EntityReference(ResourceSpec.EntityLogicalName, resourceSpecId)
                    };

                    //create the service
                    Logger.WriteDebugMessage("Creating new Service");
                    var newServiceId = OrganizationService.Create(newService);
                    UpdateTSA(thisTSA, patSites, providers, patSiteVCs, provSiteVCs, Logger, OrganizationService, srv, newServiceId);
                    #endregion
                }
                catch (FaultException<OrganizationServiceFault> ex)
                {
                    Logger.WriteToFile(ex.Message);
                    throw new InvalidPluginExecutionException(McsSettings.getUnexpectedErrorMessage + ":" + ex.Message);
                }
                catch (Exception ex)
                {
                    if (ex.Message.StartsWith("custom"))
                    {
                        Logger.WriteDebugMessage(ex.Message.Substring(6));

                        throw new InvalidPluginExecutionException(ex.Message.Substring(6));
                    }
                    else
                    {
                        Logger.setMethod = "Execute";
                        Logger.WriteToFile(ex.Message);
                        throw new InvalidPluginExecutionException(McsSettings.getUnexpectedErrorMessage + ":" + ex.Message);
                    }
                }
            }
        }

        /// <summary>
        /// Passing in the Resource Group GUID, return the StringBuilder of the associated child records.
        /// </summary>
        /// <param name="resourceGroupId"></param>
        /// <param name="srv"></param>
        /// <param name="resourceNames"></param>
        /// <param name="count"></param>
        /// <returns></returns>
        public static StringBuilder GetResources(Guid resourceGroupId, Xrm srv, out string resourceNames, out int count)
        {
            resourceNames = String.Empty;
            var resources = new StringBuilder();
            var grs = from resGroup in srv.mcs_groupresourceSet
                      join r in srv.mcs_resourceSet on resGroup.mcs_RelatedResourceId.Id equals r.mcs_resourceId.Value into resource
                      from res in resource.DefaultIfEmpty()
                      where resGroup.mcs_relatedResourceGroupId.Id == resourceGroupId && resGroup.statecode == 0
                      select new
                      {
                          name = resGroup.mcs_name,
                          user = resGroup.mcs_RelatedUserId,
                          resource = res.mcs_relatedResourceId
                      };
            count = grs.ToList().Count;

            foreach (var res in grs)
            {
                resourceNames += " " + res.name + " ;";
                if (res.user != null)
                    resources.Append(CvtHelper.AddResourceToConstraintGroup(res.user.Id.ToString()));
                else if (res.resource != null)
                    resources.Append(CvtHelper.AddResourceToConstraintGroup(res.resource.Id.ToString()));
            }
            return resources;
        }

        /// <summary>
        /// Add the Resources to the Constraint Group
        /// </summary>
        /// <param name="resourceId"></param>
        /// <returns></returns>
        public static string AddResourceToConstraintGroup(string resourceId)
        {
            var builder = "resource[\"Id\"] == ";
            builder += new Guid(resourceId).ToString("B");
            builder += " || ";
            return builder;
        }

        /// <summary>
        /// wraps the xml body string with the appropriate tags to make it a valid constraints field in a constraintbasedgroup record
        /// </summary>
        /// <param name="strBuilder">the string that becomes the body of the constraintbasedgroup</param>
        /// <returns>the full Constraints xml string</returns>
        public static StringBuilder BuildConstraintsXML(StringBuilder strBuilder)
        {
            var mainStrBuilder = new System.Text.StringBuilder("<Constraints><Constraint><Expression><Body>");
            if (strBuilder.Length == 0)
                strBuilder.Append("false || ");
            mainStrBuilder.Append(strBuilder);
            //Remove " || " from end and close the Constraint Group with the correct tags
            mainStrBuilder.Remove(mainStrBuilder.Length - 4, 4);
            mainStrBuilder.Append("</Body><Parameters><Parameter name=\"resource\" /></Parameters></Expression></Constraint></Constraints>");
            return mainStrBuilder;
        }

        /// <summary>
        /// Build Constraint Based Group
        /// </summary>
        /// <param name="thisTSA"></param>
        /// <param name="strBuilder"></param>
        /// <param name="OrgService"></param>
        /// <param name="ReqCount"></param>
        /// <param name="constraintBasedGroupTypeCode"></param>
        /// <param name="Site"></param>
        /// <returns></returns>
        public static Guid BuildOut(mcs_services thisTSA, StringBuilder strBuilder, IOrganizationService OrgService, int ReqCount, int constraintBasedGroupTypeCode, Boolean Site)
        {
            //constraintBasedGroupTypeCode
            //        Static = 0,
            //        Dynamic = 1,
            //        Implicit = 2

            //Correctly tag the XML
            var mainStrBuilder = BuildConstraintsXML(strBuilder);

            var constraintBasedGroupSetup = new ConstraintBasedGroup
            {
                BusinessUnitId = new EntityReference(BusinessUnit.EntityLogicalName, thisTSA.OwningBusinessUnit.Id),
                Constraints = mainStrBuilder.ToString(),
                Name = thisTSA.mcs_name,
                GroupTypeCode = new OptionSetValue(constraintBasedGroupTypeCode)
            };
            //Create the new Constraint Based Group
            var newConstraintGroup = OrgService.Create(constraintBasedGroupSetup);

            var newSpec = new ResourceSpec
            {
                BusinessUnitId = new EntityReference(BusinessUnit.EntityLogicalName, thisTSA.OwningBusinessUnit.Id),
                ObjectiveExpression = @"<Expression><Body>udf ""Random""(factory,resource,appointment,request,leftoffset,rightoffset)</Body><Parameters><Parameter name=""factory"" /><Parameter name=""resource"" /><Parameter name=""appointment"" /><Parameter name=""request"" /><Parameter name=""leftoffset"" /><Parameter name=""rightoffset"" /></Parameters><Properties EvaluationInterval=""P0D"" evaluationcost=""0"" /></Expression>",
                RequiredCount = ReqCount,
                Name = "Selection Rule",
                GroupObjectId = newConstraintGroup,
                SameSite = Site
            };
            var specId = OrgService.Create(newSpec);
            return specId;
        }

        /// <summary>
        /// Update TSA fields. Only update if values are different (clean up audit history)
        /// </summary>
        /// <param name="thisTSA"></param>
        /// <param name="patSites"></param>
        /// <param name="providers"></param>
        /// <param name="patSiteVCs"></param>
        /// <param name="provSiteVCs"></param>
        /// <param name="Logger"></param>
        /// <param name="OrganizationService"></param>
        /// <param name="srv"></param>
        /// <param name="newServiceId"></param>
        public static void UpdateTSA(mcs_services thisTSA, String patSites, String providers, String patSiteVCs, String provSiteVCs, MCSLogger Logger, IOrganizationService OrganizationService, Xrm srv, Guid newServiceId)
        {
            Logger.setMethod = "Update TSA";
            Logger.WriteDebugMessage("About to update TSA");
            var updateTSA = new Entity("mcs_services") { Id = thisTSA.Id };
            var updateCount = 0;
            Guid formerService = Guid.Empty;

            if (newServiceId == Guid.Empty) //Draft
            {
                if (thisTSA.mcs_RelatedServiceId != null)
                {
                    //Set to null
                    formerService = thisTSA.mcs_RelatedServiceId.Id;
                    updateTSA["mcs_relatedserviceid"] = null;
                    updateCount += 1;
                }
            }
            else //Production - always use the new id if present
            {
                //Check if there is a former value in service
                if (thisTSA.mcs_RelatedServiceId != null)
                    formerService = thisTSA.mcs_RelatedServiceId.Id;
                updateTSA["mcs_relatedserviceid"] = new EntityReference(Service.EntityLogicalName, newServiceId);
                updateCount += 1;
            }

            if (thisTSA.cvt_patientsites != ValidateLength(patSites, 2500))
            {
                updateTSA["cvt_patientsites"] = ValidateLength(patSites, 2500);
                updateCount += 1;
            }
            if (thisTSA.cvt_providers != ValidateLength(providers, 2500))
            {
                updateTSA["cvt_providers"] = ValidateLength(providers, 2500);
                updateCount += 1;
            }
            if (thisTSA.cvt_patsitevistaclinics != ValidateLength(patSiteVCs, 2500))
            {
                updateTSA["cvt_patsitevistaclinics"] = ValidateLength(patSiteVCs, 2500);
                updateCount += 1;
            }
            if (thisTSA.cvt_provsitevistaclinics != ValidateLength(provSiteVCs, 2500))
            {
                updateTSA["cvt_provsitevistaclinics"] = ValidateLength(provSiteVCs, 2500);
                updateCount += 1;
            }

            //Clear Bulk Edit Fields
            if (thisTSA.cvt_BulkRemoveResource != null)
            {
                updateTSA["cvt_bulkremoveresource"] = null;
                updateCount += 1;
            }
            if (thisTSA.cvt_BulkAddResource != null)
            {
                updateTSA["cvt_bulkaddresource"] = null;
                updateCount += 1;
            }
            if (thisTSA.cvt_BulkRemoveUser != null)
            {
                updateTSA["cvt_bulkremoveuser"] = null;
                updateCount += 1;
            }
            if (thisTSA.cvt_BulkAddUser != null)
            {
                updateTSA["cvt_bulkadduser"] = null;
                updateCount += 1;
            }
            if (thisTSA.cvt_BulkRemoveResourceGroup != null)
            {
                updateTSA["cvt_bulkremoveresourcegroup"] = null;
                updateCount += 1;
            }
            if (thisTSA.cvt_BulkAddResourceGroup != null)
            {
                updateTSA["cvt_bulkaddresourcegroup"] = null;
                updateCount += 1;
            }

            if (updateCount > 0)
            {
                OrganizationService.Update(updateTSA);
                Logger.WriteDebugMessage("TSA Updated");
                if (formerService != Guid.Empty)
                {
                    //Search on the prior service and attempt to delete it.
                    var ServiceActivity = srv.ServiceAppointmentSet.FirstOrDefault(i => i.ServiceId.Id == formerService);
                    //If an associated SA doesn't exist, delete
                    if (ServiceActivity == null)
                    {
                        OrganizationService.Delete(Service.EntityLogicalName, formerService);
                        Logger.WriteDebugMessage("Prior Service successfully deleted.");
                    }
                    else
                        Logger.WriteDebugMessage("Prior Service could not be deleted, associated Service Activities exist.");
                }
            }
            else
                Logger.WriteDebugMessage("No systematically built values to update on TSA");
        }

        /// <summary>
        /// Update MTSA fields
        /// </summary>
        /// <param name="mtsaId"></param>
        /// <param name="deleteId"></param>
        /// <param name="Logger"></param>
        /// <param name="OrganizationService"></param>
        public static void UpdateMTSA(Guid mtsaId, Guid deleteId, MCSLogger Logger, IOrganizationService OrganizationService)
        {
            Logger.setMethod = "UpdateMTSA";
            Logger.WriteDebugMessage("starting UpdateMTSA");
            using (var srv = new Xrm(OrganizationService))
            {
                if (mtsaId == Guid.Empty)
                {
                    Logger.WriteDebugMessage("No MTSA value, exiting.");
                    return;
                }

                var getProvResources = from provGroups in srv.cvt_providerresourcegroupSet
                                        where provGroups.cvt_RelatedMasterTSAId.Id == mtsaId
                                        where provGroups.statecode == 0
                                        select new
                                        {
                                            provGroups.Id,
                                            provGroups.cvt_RelatedResourceId,
                                            provGroups.cvt_resourcespecguid,
                                            provGroups.cvt_TSAResourceType,
                                            provGroups.cvt_Type,
                                            provGroups.cvt_RelatedUserId,
                                            provGroups.cvt_RelatedResourceGroupid
                                        };

                //grab all the provider groups
                string provSiteVCs = null;
                string providers = null;
                foreach (var provGroups in getProvResources)
                {
                    if (provGroups.Id != deleteId)
                    {
                        //Verify that the Resource is typed, not required, but should be filled in
                        if (provGroups.cvt_Type != null)
                        {
                            //Naming
                            switch (provGroups.cvt_Type.Value)
                            {
                                case 251920000: //Vista Clinic
                                    if (provGroups.cvt_TSAResourceType.Value == 0) //Group of Vista Clinics
                                    {
                                        //Query for child names
                                        var groupResourceRecords = srv.mcs_groupresourceSet.Where(g => g.mcs_relatedResourceGroupId.Id == provGroups.cvt_RelatedResourceGroupid.Id);
                                        foreach (var child in groupResourceRecords)
                                        {
                                            if (child.mcs_RelatedResourceId != null)
                                                provSiteVCs += child.mcs_RelatedResourceId.Name.ToString() + " ; ";
                                        }
                                    }
                                    else
                                    {
                                        if (provGroups.cvt_RelatedResourceId != null)
                                            provSiteVCs += provGroups.cvt_RelatedResourceId.Name.ToString() + " ; ";
                                    }
                                    break;
                                case 100000000: //Telepresenter/Imager
                                case 99999999: //Provider
                                    if (provGroups.cvt_TSAResourceType.Value == 0) //Group of Providers
                                    {
                                        //Query for child names
                                        var groupResourceRecords = srv.mcs_groupresourceSet.Where(g => g.mcs_relatedResourceGroupId.Id == provGroups.cvt_RelatedResourceGroupid.Id);
                                        foreach (var child in groupResourceRecords)
                                        {
                                            if (child.mcs_RelatedUserId != null)
                                                providers += child.mcs_RelatedUserId.Name.ToString() + " ; ";
                                        }
                                    }
                                    else
                                    {
                                        if (provGroups.cvt_RelatedUserId != null)
                                            providers += provGroups.cvt_RelatedUserId.Name.ToString() + " ; ";
                                    }
                                    break;
                                case 917290000: //All Required
                                    //Query for child names
                                    var childgroupResourceRecords = srv.mcs_groupresourceSet.Where(g => g.mcs_relatedResourceGroupId.Id == provGroups.cvt_RelatedResourceGroupid.Id);
                                    foreach (var child in childgroupResourceRecords)
                                    {
                                        if (child.mcs_RelatedUserId != null)
                                            providers += child.mcs_RelatedUserId.Name.ToString() + " ; ";
                                        else
                                        { //Check the related Resource if it is a Vista Clinic
                                            var childR = srv.mcs_resourceSet.First(r => r.Id == child.mcs_RelatedResourceId.Id);
                                            if (childR.mcs_Type.Value == 251920000)
                                                provSiteVCs += childR.mcs_name + " ; ";
                                        }
                                    }
                                    break;
                            }
                        }
                        else //Probably Single Provider, but check.
                        {
                            //Provider or Telepresenter
                            if ((provGroups.cvt_TSAResourceType.Value == 2) || (provGroups.cvt_TSAResourceType.Value == 3))
                            {
                                if (provGroups.cvt_RelatedUserId != null)
                                {
                                    providers += provGroups.cvt_RelatedUserId.Name.ToString() + " ; ";
                                }
                            }
                        }                   
                    }
                }
                
                var mtsa = srv.cvt_mastertsaSet.FirstOrDefault(i => i.Id == mtsaId);
                var updateMTSA = new Entity("cvt_mastertsa") { Id = mtsaId };
                var updateCount = 0;

                if (mtsa.cvt_providers != ValidateLength(providers, 2500))
                {
                    updateMTSA["cvt_providers"] = ValidateLength(providers, 2500);
                    updateCount += 1;
                }
                if (mtsa.cvt_providersitevistaclinics != ValidateLength(provSiteVCs, 2500))
                {
                    updateMTSA["cvt_providersitevistaclinics"] = ValidateLength(provSiteVCs, 2500);
                    updateCount += 1;
                }
                if (updateCount > 0)
                {
                    OrganizationService.Update(updateMTSA);
                    Logger.WriteDebugMessage("MTSA Updated");
                }
                else
                    Logger.WriteDebugMessage("No systematically built values to update on MTSA");
            }
        }

        /// <summary>
        /// Check length of a string and either reduce or return the original value
        /// </summary>
        /// <param name="value"></param>
        /// <param name="length"></param>
        /// <returns></returns>
        public static String ValidateLength(String value, int length)
        {
            if (value != null)
            {
                if (value.Length > length)
                    return value.Substring(0, length);
            }
            return value;
        }
        #endregion

        #region Calendar Methods
        //FROM CREATE
        /// <summary>
        /// This method is used to update the calendar for the equipment record passed in.  It uses a combination of the fields passed in and the values from the default calendar.  
        /// </summary>
        /// <param name="sysResourceId">The ID of the equipment record corresponding to the calendar to be updated</param>
        /// <param name="OrganizationService">The OrganizationService from the plugin execution context</param>
        /// <param name="Logger">The Logger from the plugin execution context</param>
        /// <param name="McsSettings">The McsSettings from the plugin execution context</param>
        internal static void ChangeNewlyCreatedCalendar(Guid sysResourceId, IOrganizationService OrganizationService, MCSLogger Logger, MCSSettings McsSettings)
        {
            Logger.setMethod = "ChangeNewlyCreatedCalendar";
            Logger.WriteDebugMessage("starting ChangeNewlyCreatedCalendar");
            using (var srv = new Xrm(OrganizationService))
            {
                try
                {
                    var newTimeCode = 0;
                    var i = 0;
                    var thisEquipment = srv.EquipmentSet.FirstOrDefault(e => e.Id == sysResourceId);

                    if (thisEquipment == null)
                        return;
                    Logger.WriteDebugMessage("Equipment name is " + thisEquipment.Name);
                    var defaultEquipmentfromSettings = McsSettings.GetSingleSetting("mcs_defaultcalendar", "entityreference");
                    Logger.WriteDebugMessage("Got EquipmentId");
                    var defaultEquipmentId = new Guid(defaultEquipmentfromSettings);

                    var defaultEquipment = srv.EquipmentSet.FirstOrDefault(u => u.EquipmentId == defaultEquipmentId);
                    if (defaultEquipment == null)
                        return;

                    Logger.WriteDebugMessage("Got Default Equipment");
                    Entity defaultCalendarEntity = OrganizationService.Retrieve("calendar", defaultEquipment.CalendarId.Id, new ColumnSet(true));
                    Logger.WriteDebugMessage("Got default Calendar");
                    Entity thisCalendarEntity = OrganizationService.Retrieve("calendar", thisEquipment.CalendarId.Id, new ColumnSet(true));
                    Logger.WriteDebugMessage("Got This Calendar");

                    // Retrieve the calendar rules defined in the calendar
                    EntityCollection thiscalendarRules = (EntityCollection)thisCalendarEntity.Attributes["calendarrules"];
                    EntityCollection defaultCalendarRules = (EntityCollection)defaultCalendarEntity.Attributes["calendarrules"];

                    foreach (CalendarRule defaultCalRule in defaultCalendarRules.Entities)
                    {
                        Logger.WriteDebugMessage("Starting new Loop: " + i);
                        Entity defaultInnerCalendarEntity = OrganizationService.Retrieve("calendar", defaultCalRule.InnerCalendarId.Id, new ColumnSet(true));
                        EntityCollection defaultInnerCalendarRules = (EntityCollection)defaultInnerCalendarEntity.Attributes["calendarrules"];

                        if (thiscalendarRules.Entities.Count >= (i + 1))
                        {
                            Logger.WriteDebugMessage("Creating new calendar:" + i);
                            CalendarRule newRule = (thiscalendarRules.Entities[i]).ToEntity<CalendarRule>();
                            newRule.Duration = defaultCalRule.Duration;
                            newRule.EffectiveIntervalEnd = defaultCalRule.EffectiveIntervalEnd;
                            newRule.EffectiveIntervalStart = defaultCalRule.EffectiveIntervalStart;
                            newRule.StartTime = defaultCalRule.StartTime;
                            newRule.Pattern = defaultCalRule.Pattern;
                            
                            ///Adding Effort
                            if (thisEquipment.cvt_type.Value == 251920000 && thisEquipment.cvt_capacity != null)
                                newRule.Effort = thisEquipment.cvt_capacity.Value;

                            newTimeCode = newRule.TimeZoneCode.Value;

                            if (i == 0)
                            {
                                Logger.WriteDebugMessage("First Loop");
                                if (defaultCalRule.InnerCalendarId != null)
                                {
                                    CalendarRule defaultInnerCalRule = new CalendarRule();
                                    if ((defaultInnerCalendarRules.Entities != null) && (defaultInnerCalendarRules.Entities.Count > 0))
                                    {
                                         defaultInnerCalRule = (defaultInnerCalendarRules.Entities[0]).ToEntity<CalendarRule>();
                                    }
                                    else
                                    {
                                        Logger.WriteDebugMessage("Could not find the Default Calendar's Inner Calendar Rule. Ref: CvtHelper.ChangeNewlyCreatedCalendar");
                                        break;
                                    }
                                    Entity thisInnerCalendarEntity = OrganizationService.Retrieve("calendar", newRule.InnerCalendarId.Id, new ColumnSet(true));
                                    EntityCollection thisInnerCalendarRules = (EntityCollection)thisInnerCalendarEntity.Attributes["calendarrules"];
                                    CalendarRule thisInnerCalRule = new CalendarRule();
                                    
                                    //there is always 1 and only 1 inner calendar, but there may be many more than 1 outer calendar. 
                                    //Update: There are instances where there isn't 1 inner calendar. Need to check
                                    if ((thisInnerCalendarRules.Entities != null) && (thisInnerCalendarRules.Entities.Count > 0))
                                        thisInnerCalRule = (thisInnerCalendarRules.Entities[0]).ToEntity<CalendarRule>();

                                    else //Figure out what the other properties of thisInnterCalendarRules' CalendarRule, and set them.
                                    {
                                        thisInnerCalRule = defaultInnerCalRule;
                                        thisInnerCalRule.CalendarId.Id = thisInnerCalendarEntity.Id;  
                                    }
                                   
                                    if (thisEquipment.cvt_type.Value == 251920000 && thisEquipment.cvt_capacity != null)
                                        thisInnerCalRule.Effort = thisEquipment.cvt_capacity.Value;
                                    else
                                        thisInnerCalRule.Effort = defaultInnerCalRule.Effort;

                                    thisInnerCalRule.Offset = defaultInnerCalRule.Offset;
                                    thisInnerCalRule.Duration = defaultInnerCalRule.Duration;

                                    if (thisInnerCalRule.TimeZoneCode == null)
                                        thisInnerCalRule.TimeZoneCode = newTimeCode;
                                    if (thisInnerCalRule.Rank == null)
                                        thisInnerCalRule.Rank = defaultInnerCalRule.Rank == null ? 0 : defaultInnerCalRule.Rank;

                                    if (thisInnerCalRule.InnerCalendarId == null)
                                        thisInnerCalRule.IsSimple = true;
                                    else
                                        thisInnerCalRule.IsSimple = null;

                                    Logger.WriteDebugMessage(String.Format("Duration: {0}, Rank: {1}, Time Zone: {2}", thisInnerCalRule.Duration, thisInnerCalRule.Rank, thisInnerCalRule.TimeZoneCode));
                                    OrganizationService.Update(thisInnerCalendarEntity);
                                    Logger.WriteDebugMessage("Updating Calendar");
                                }
                            }
                        }
                        else
                        {
                            //this means that the default calendar has more than 1 calendar entry.
                            if (defaultCalRule.Name == "Holiday Closure Link")
                            {
                                Logger.WriteDebugMessage("Beginning Holiday Closure Link Default Calendar");
                                //this means the resource is following business closures.  Copy EVERYTHING from this one.
                                CalendarRule newRule = new CalendarRule();
                                newRule.Rank = defaultCalRule.Rank;
                                newRule.StartTime = defaultCalRule.StartTime;
                                newRule.Name = defaultCalRule.Name;
                                newRule.TimeZoneCode = newTimeCode;
                                newRule.ExtentCode = defaultCalRule.ExtentCode;
                                newRule.EffectiveIntervalEnd = defaultCalRule.EffectiveIntervalEnd;
                                newRule.Pattern = defaultCalRule.Pattern;
                                newRule.InnerCalendarId = defaultCalRule.InnerCalendarId;
                                newRule.GroupDesignator = defaultCalRule.GroupDesignator;
                                newRule.Duration = defaultCalRule.Duration;
                                newRule.Description = "Default";
                                //Adding Effort
                                if (thisEquipment.cvt_type.Value == 251920000 && thisEquipment.cvt_capacity != null)
                                    newRule.Effort = thisEquipment.cvt_capacity.Value;

                                if (newRule.InnerCalendarId == null)
                                    newRule.IsSimple = true;
                                else
                                    newRule.IsSimple = null;

                                thiscalendarRules.Entities.Add(newRule);
                                Logger.WriteDebugMessage("Updating Calendar to include holidays");
                            }
                            else
                            {
                                Logger.WriteDebugMessage("New Calendar, not Holiday Closure Link");
                                // Create a new  calendar
                                Entity newCalendar = new Entity("calendar");
                                newCalendar.Attributes["businessunitid"] = new EntityReference("businessunit", ((Microsoft.Xrm.Sdk.EntityReference)(thisCalendarEntity["businessunitid"])).Id);
                                newCalendar.Attributes["description"] = defaultCalendarEntity["description"];
                                Guid innerCalendarId = OrganizationService.Create(newCalendar);

                                // Create a new calendar rule and assign the inner calendar id to it
                                CalendarRule calendarRule = new CalendarRule();
                                calendarRule.ExtentCode = defaultCalRule.ExtentCode;
                                calendarRule.StartTime = defaultCalRule.StartTime;
                                calendarRule.TimeZoneCode = newTimeCode;
                                calendarRule.Pattern = defaultCalRule.Pattern;
                                calendarRule.Duration = defaultCalRule.Duration;
                                calendarRule.Rank = defaultCalRule.Rank;
                                calendarRule.InnerCalendarId = new EntityReference("calendar", innerCalendarId);

                                //Adding Effort
                                if (thisEquipment.cvt_type.Value == 251920000 && thisEquipment.cvt_capacity != null)
                                    calendarRule.Effort = thisEquipment.cvt_capacity.Value;

                                if (calendarRule.InnerCalendarId == null)
                                    calendarRule.IsSimple = true;
                                else
                                    calendarRule.IsSimple = null;

                                thiscalendarRules.Entities.Add(calendarRule);
                                Logger.WriteDebugMessage("Adding new innerCalendar");

                                if (defaultCalRule.InnerCalendarId != null)
                                {
                                    CalendarRule defaultInnerCalRule = new CalendarRule();
                                    //there is always 1 and only 1 inner calendar, but there may be many more than 1 outer calendar.
                                    if ((defaultInnerCalendarRules.Entities != null) && (defaultInnerCalendarRules.Entities.Count > 0))
                                    {
                                        defaultInnerCalRule = (defaultInnerCalendarRules.Entities[0]).ToEntity<CalendarRule>();
                                    }
                                    else
                                    {
                                        Logger.WriteDebugMessage("Could not find the Default Calendar's Inner Calendar Rule. Ref: CvtHelper.ChangeNewlyCreatedCalendar");
                                        break;
                                    }

                                    CalendarRule newRule = new CalendarRule();
                                    newRule.Description = "Default";
                                    newRule.Rank = defaultInnerCalRule.Rank;
                                    newRule.StartTime = defaultInnerCalRule.StartTime;
                                    newRule.Name = defaultInnerCalRule.Name;
                                    newRule.TimeZoneCode = newTimeCode;
                                    newRule.Offset = defaultInnerCalRule.Offset;
                                    newRule.ExtentCode = defaultInnerCalRule.ExtentCode;
                                    newRule.EffectiveIntervalEnd = defaultInnerCalRule.EffectiveIntervalEnd;
                                    newRule.IsSimple = defaultInnerCalRule.IsSimple;
                                    newRule.GroupDesignator = defaultInnerCalRule.GroupDesignator;
                                    newRule.Duration = defaultInnerCalRule.Duration;
                                    newRule.TimeCode = defaultInnerCalRule.TimeCode;
                                    newRule.SubCode = defaultInnerCalRule.SubCode;

                                    //Adding Effort
                                    if (thisEquipment.cvt_type.Value == 251920000 && thisEquipment.cvt_capacity != null)
                                        newRule.Effort = thisEquipment.cvt_capacity.Value;

                                    EntityCollection innerCalendarRules = new EntityCollection();
                                    innerCalendarRules.EntityName = "calendarrule";
                                    innerCalendarRules.Entities.Add(newRule);

                                    newCalendar.Attributes["calendarrules"] = innerCalendarRules;
                                    newCalendar.Attributes["calendarid"] = innerCalendarId;

                                    Logger.WriteDebugMessage("Updating Inner Service Calendar " + i);
                                    OrganizationService.Update(newCalendar);
                                    Logger.WriteDebugMessage("Service Update Inner");
                                }
                            }
                        }
                        i += 1;
                    }
                    //((EntityCollection)(thisCalendarEntity.Attributes["calendarrules"])).Entities.Count));
                    Logger.WriteDebugMessage(String.Format("Attempting to update calendar entity with {0} rules", ((EntityCollection)(thisCalendarEntity.Attributes["calendarrules"])).Entities.Count));
                    OrganizationService.Update(thisCalendarEntity);
                    Logger.WriteDebugMessage("Service Update Outer");
                }
                catch (FaultException<OrganizationServiceFault> ex)
                {
                    Logger.WriteToFile(ex.Message);
                    throw new InvalidPluginExecutionException(ex.Message);
                }
                catch (Exception ex)
                {
                    Logger.WriteToFile(ex.Message);
                    throw new InvalidPluginExecutionException(ex.Message);
                }
            }
        }

        #endregion

        #region Denormalized Data Integrity Functions

        /// <summary>
        /// OVERLOAD: Updated TSS Resource record - with a new TSS Site
        /// </summary>
        /// <param name="tssresource"></param>
        /// <param name="OrganizationService"></param>
        /// <param name="Logger"></param>
        internal static void AlignLocations(mcs_resource tssresource, IOrganizationService OrganizationService, MCSLogger Logger)
        {
            //Update this one TSS Resources - Facility based on Site, and check VISN based on Facility
            if (tssresource.mcs_RelatedSiteId != null)
                AlignOneTSSResource(tssresource.Id, OrganizationService, Logger);
        }
        
        /// <summary>
        /// OVERLOAD: Updated TSS Site record.  Can only update the Facility on this record.
        /// </summary>
        /// <param name="site"></param>
        /// <param name="OrganizationService"></param>
        /// <param name="Logger"></param>
        internal static void AlignLocations(mcs_site site, IOrganizationService OrganizationService, MCSLogger Logger)
        {
            //Update TSS Sites - VISN based on Facility -Possibly only update 1, if the site changed facility that is in another visn
            if (site.mcs_FacilityId != null)
                AlignTSSSiteVISNData(site.mcs_FacilityId.Id, OrganizationService, Logger);

            //Update Users - Facility based on Site
            AlignUserFacilityData(site.Id, OrganizationService, Logger);

            //Update TSS Resources - Facility based on Site, VISN based on Facility
            AlignMultipleTSSResourceFacilityVISNData(site.Id, OrganizationService, Logger);           
        }
    
        //New Plugin for Facility Update and step
        /// <summary>
        /// OVERLOAD: Updated Facility record - with a new VISN
        /// </summary>
        /// <param name="facility"></param>
        /// <param name="OrganizationService"></param>
        /// <param name="Logger"></param>
        internal static void AlignLocations(mcs_facility facility, IOrganizationService OrganizationService, MCSLogger Logger)
        {
            //Update TSS Sites - VISN based on Facility
            AlignTSSSiteVISNData(facility.Id, OrganizationService, Logger);

            //Update TSS Resources - Facility based on Site and VISN based on Facility
            AlignMultipleTSSResourceFacilityVISNData(facility.Id, OrganizationService, Logger);

            //AlignSystemUserVISNData(facility.Id, OrganizationService, Logger);
        }

        /// <summary>
        /// TSS Site Entity
        /// Align VISN from Facility
        /// </summary>
        /// <param name="thisFacilityId"></param>
        /// <param name="OrganizationService"></param>
        /// <param name="Logger"></param>
        internal static void AlignTSSSiteVISNData(Guid thisFacilityId, IOrganizationService OrganizationService, MCSLogger Logger)
        {
            using (var srv = new Xrm(OrganizationService))
            {
                Logger.setMethod = "AlignTSSSiteVISNData";
                Logger.WriteTxnTimingMessage("starting AlignTSSSiteVISNData");
                var thisFacility = srv.mcs_facilitySet.FirstOrDefault(i => i.Id == thisFacilityId);
                var countUpdate = 0;

                if (thisFacility.mcs_BusinessUnitId != null)
                {

                    var retrieveTSSSites = from s in srv.mcs_siteSet
                                           where s.mcs_FacilityId.Id == thisFacilityId
                                               select new
                                               {
                                                   s.Id,
                                                   s.mcs_BusinessUnitId,
                                                   s.mcs_name
                                               };
                    
                
                    //Loop through all of the TSS Sites with this VISN
                    foreach (var tsssites in retrieveTSSSites)
                    {
                        var update = false;
                        var updateThisTSSSite = new mcs_site
                        {
                            Id = tsssites.Id
                        };

                        if ((tsssites.mcs_BusinessUnitId == null) || (thisFacility.mcs_BusinessUnitId.Id != tsssites.mcs_BusinessUnitId.Id))
                        {
                            //Need to update TSS Resources Facility
                            updateThisTSSSite.mcs_BusinessUnitId = thisFacility.mcs_BusinessUnitId;
                            update = true;
                        }

                        if (update)
                        {
                            OrganizationService.Update(updateThisTSSSite);
                            countUpdate += 1;
                            Logger.WriteTxnTimingMessage(tsssites.mcs_name + "'s VISN was updated");
                        }
                    }
                }
                Logger.WriteTxnTimingMessage("ending AlignTSSSiteVISNData.");
                Logger.WriteDebugMessage(countUpdate + " TSS Sites' VISN were updated.");   
            }
        }

        /// <summary>
        /// Users Entity
        /// Align Facility from Site
        /// </summary>
        /// <param name="thisSiteId"></param>
        /// <param name="OrganizationService"></param>
        /// <param name="Logger"></param>
        internal static void AlignUserFacilityData(Guid thisSiteId, IOrganizationService OrganizationService, MCSLogger Logger)
        {
            using (var srv = new Xrm(OrganizationService))
            {
                Logger.setMethod = "AlignUserFacilityData";
                Logger.WriteTxnTimingMessage("starting AlignUserFacilityData");
                //Need to retrieve Site again, because it has changed
                var thisUpdatedSite = srv.mcs_siteSet.FirstOrDefault(i => i.Id == thisSiteId);
                var retrieveUsers = from u in srv.SystemUserSet
                                    where u.cvt_site.Id == thisSiteId
                                    select new
                                    {
                                        u.Id,
                                        u.cvt_facility,
                                        u.FullName
                                    };
                var countUpdate = 0;
                //Retrieve all Users with this TSS Site
                foreach (var users in retrieveUsers)
                {
                    var update = false;
                    var updateThisUser = new SystemUser
                    {
                        Id = users.Id
                    };
                    if (thisUpdatedSite.mcs_FacilityId != null)
                    {

                        if ((users.cvt_facility == null) || (thisUpdatedSite.mcs_FacilityId.Id != users.cvt_facility.Id))
                        {
                            //Need to update TSS Resources Facility
                            updateThisUser.cvt_facility = thisUpdatedSite.mcs_FacilityId;
                            update = true;
                        }
                    }
                    if (update == true)
                    {
                        OrganizationService.Update(updateThisUser);
                        countUpdate += 1;
                        Logger.WriteTxnTimingMessage(users.FullName + " Facility was updated based on Site");
                    }
                }
                Logger.WriteTxnTimingMessage("ending AlignUserFacilityData.");
                Logger.WriteDebugMessage(countUpdate + " Users were updated.");
            }
        }


        internal static int AlignOneTSSResource(Guid thistssresourceId, IOrganizationService OrganizationService, MCSLogger Logger)
        {
            using (var srv = new Xrm(OrganizationService))
            {
                Logger.setMethod = "AlignOneTSSResource";
                Logger.WriteTxnTimingMessage("starting AlignOneTSSResource");
                var thisresource = srv.mcs_resourceSet.FirstOrDefault(i => i.Id == thistssresourceId);
                var resourcesSite = srv.mcs_siteSet.FirstOrDefault(i => i.Id == thisresource.mcs_RelatedSiteId.Id);
                var sitesFacility = srv.mcs_facilitySet.FirstOrDefault(i => i.Id == resourcesSite.mcs_FacilityId.Id);
                Logger.WriteDebugMessage("Retrieved Resource, Site, Facility");

                //Create a tss resource object and count
                var update = false;
                var updateThisTSSResource = new mcs_resource
                {
                    Id = thistssresourceId
                };

                //Compare the Resource's Facility to the Site's Facility
                if ((thisresource.mcs_Facility == null) || (thisresource.mcs_Facility.Id != resourcesSite.mcs_FacilityId.Id))
                {
                    updateThisTSSResource.mcs_Facility = resourcesSite.mcs_FacilityId;
                    update = true;
                    Logger.WriteDebugMessage("Need to update this Resource's Facility");
                }
                //Compare the Resource's VISN to the Facility's VISN
                if ((thisresource.mcs_BusinessUnitId == null) || (thisresource.mcs_BusinessUnitId.Id != sitesFacility.mcs_BusinessUnitId.Id))
                {
                    updateThisTSSResource.mcs_BusinessUnitId = sitesFacility.mcs_BusinessUnitId;
                    update = true;
                    Logger.WriteDebugMessage("Need to update this Resource's VISN");
                }

                //Update record if anything is different
                if (update)
                {
                    OrganizationService.Update(updateThisTSSResource);
                    Logger.WriteDebugMessage("Updated TSS Resource");
                    return 1;
                }
                else
                    return 0;
            }
        }
        /// <summary>
        /// TSS Resource Entity
        /// Align Facility from Site
        /// Calls AlignTSSResourceVISNData at end
        /// </summary>
        /// <param name="thisSiteId"></param>
        /// <param name="OrganizationService"></param>
        /// <param name="Logger"></param>
        internal static void AlignMultipleTSSResourceFacilityVISNData(Guid thisSiteId, IOrganizationService OrganizationService, MCSLogger Logger)
        {
            using (var srv = new Xrm(OrganizationService))
            {
                Logger.setMethod = "AlignMultipleTSSResourceFacilityVISNData";
                Logger.WriteTxnTimingMessage("starting AlignMultipleTSSResourceFacilityVISNData");
                //Need to retrieve Site again, because it has changed
                var thisUpdatedSite = srv.mcs_siteSet.FirstOrDefault(i => i.Id == thisSiteId);
                var retrieveTSSResources = from r in srv.mcs_resourceSet
                                           where r.mcs_RelatedSiteId.Id == thisSiteId
                                           select new
                                           {
                                               r.Id,
                                               r.mcs_Facility,
                                               r.mcs_name
                                           };
                var countUpdate = 0;
                //Loop through all of the TSS resources with this site
                foreach (var tssresource in retrieveTSSResources)
                {
                    countUpdate += AlignOneTSSResource(tssresource.Id, OrganizationService, Logger);
                }
                Logger.setMethod = "AlignMultipleTSSResourceFacilityVISNData";
                Logger.WriteTxnTimingMessage("ending AlignTSSResourceFacilityData.");
                Logger.WriteDebugMessage(countUpdate + " TSS Resources' records were updated.");             
            }
        }
        
        /// <summary>
        /// TSS Resource Entity
        /// Align VISN from Facility
        /// </summary>
        /// <param name="thisFacilityId"></param>
        /// <param name="OrganizationService"></param>
        /// <param name="Logger"></param>
        internal static void AlignTSSResourceVISNData(Guid thisFacilityId, IOrganizationService OrganizationService, MCSLogger Logger)
        {
            using (var srv = new Xrm(OrganizationService))
            {
                Logger.setMethod = "AlignTSSResourceVISNData";
                Logger.WriteTxnTimingMessage("starting AlignTSSResourceVISNData");
                //Need to retrieve Site again, because it has changed
                var thisFacility = srv.mcs_facilitySet.FirstOrDefault(i => i.Id == thisFacilityId);
                var retrieveTSSResources = from r in srv.mcs_resourceSet
                                           where r.mcs_Facility.Id == thisFacilityId
                                           select new
                                           {
                                               r.Id,
                                               r.mcs_BusinessUnitId,
                                               r.mcs_name
                                           };
                var countUpdate = 0;
                //Loop through all of the TSS resources with this facility
                foreach (var tssresource in retrieveTSSResources)
                {
                    var update = false;
                    var updateThisTSSResource = new mcs_resource
                    {
                        Id = tssresource.Id
                    };

                    if (thisFacility.mcs_BusinessUnitId != null)
                    {
                        if ((tssresource.mcs_BusinessUnitId == null) || (thisFacility.mcs_BusinessUnitId.Id != tssresource.mcs_BusinessUnitId.Id))
                        {
                            //Need to update TSS Resources Facility
                            updateThisTSSResource.mcs_BusinessUnitId = thisFacility.mcs_BusinessUnitId;
                            update = true;
                        }
                    }
                    if (update == true)
                    {
                        OrganizationService.Update(updateThisTSSResource);
                        countUpdate += 1;
                        Logger.WriteTxnTimingMessage(tssresource.mcs_name + "'s VISN was updated");
                    }
                }
                Logger.WriteTxnTimingMessage("ending AlignTSSResourceVISNData.");
                Logger.WriteDebugMessage(countUpdate + " TSS Resources' VISN were updated.");
            }
        }
        #endregion

        #region Data Cleanup
        /// <summary>
        /// Passing in either a SystemUser or a Team, associate security roles
        /// </summary>
        /// <param name="thisId">Accepts either Team or SystemUser Guids</param>
        /// <param name="EntityLogicalName"></param>
        /// <param name="OrganizationService"></param>
        /// <param name="Logger"></param>
        internal static void UpdateSecurityRoles(Guid thisId, string EntityLogicalName, IOrganizationService OrganizationService, MCSLogger Logger, out int addRoles, out int removeRoles)
        {
            using (var srv = new Xrm(OrganizationService))
            {
                //Vars to Track Counts
                bool addedRoles = false;
                bool removedRoles = false;
                addRoles = 0;
                removeRoles = 0;

                IQueryable currentRoleResults = null;
                string rolesToAdd = "";
                int totalRoles = 0;
                Guid businessUnitId = Guid.Empty;
                string thisRecordName = "";
                string displayMessage = "";
                EntityReferenceCollection RolesToAssociate = new EntityReferenceCollection();
                EntityReferenceCollection RolesToDisassociate = new EntityReferenceCollection();

                if (EntityLogicalName == Team.EntityLogicalName)
                {
                    var thisTeam = srv.TeamSet.First(u => u.Id == thisId);
                    if (thisTeam.BusinessUnitId == null)
                    {
                        Logger.WriteDebugMessage("No BU set on Team, so no roles were added.");
                        //Maybe check For Facility value and align to that BU.
                        return; //no BU, so exit                      
                    }
                    businessUnitId = thisTeam.BusinessUnitId.Id;
                    thisRecordName = thisTeam.Name;
                    rolesToAdd = "TSS User";
                    if (thisTeam.cvt_Type != null)
                    {
                        switch (thisTeam.cvt_Type.Value)
                        {
                            case 917290000: //FTC
                                //User + Approver + FTC
                                rolesToAdd += "|TSS TSA Approvers|TSS FTC";
                                break;
                            case 917290001: //Service Chief
                                rolesToAdd += "|TSS TSA Approvers|TSS Privileging - Service Chief Team";
                                break;
                            case 917290002: //Chief of Staff
                                rolesToAdd += "|TSS TSA Approvers";
                                break;
                            case 917290003: //Credentialing and Privileging
                                rolesToAdd += "|TSS TSA Approvers|TSS Privileging - Privileging Team";
                                break;
                            //case "917290004": //TSA Notification
                            //    break;
                            case 917290005: //Scheduler
                                //User + Scheduler
                                rolesToAdd += "|TSS Scheduler";
                                break;
                        }
                    }

                    displayMessage = "Missing Default Role(s) systematically associated (based on Type) to Team: ";
                    currentRoleResults = srv.TeamRolesSet.Where(tr => tr.TeamId == thisId);
                }
                else if (EntityLogicalName == SystemUser.EntityLogicalName)
                {
                    var thisUser = srv.SystemUserSet.First(u => u.Id == thisId);
                    if (thisUser.cvt_SecurityRolesString == null)
                        return; //Future: Auto add Roles based on Type field on User Record (like Team)
                    rolesToAdd = thisUser.cvt_SecurityRolesString;
                    businessUnitId = thisUser.BusinessUnitId.Id;
                    thisRecordName = thisUser.FullName;
                    displayMessage = " Former Roles re-associated in new Business Unit for user: ";                  
                    currentRoleResults = srv.SystemUserRolesSet.Where(u => u.SystemUserId == thisId);                   
                }

                //Common Code
                if (rolesToAdd == "")
                {
                    //Check for VA Site Team, VA Facility Team, VA VISN Team.
                    //Could query for site or facility or visn for exact match of name and if so, add team depending on which result?
                    Logger.WriteDebugMessage(String.Format("No roles to assign for {0}: {1}", EntityLogicalName, thisRecordName));
                    return;
                }
                var SecRoles = rolesToAdd.Contains('|') ? rolesToAdd.Split('|') : new string[] { rolesToAdd }; //split or only one value
                totalRoles = SecRoles.Count();

                try
                {
                    //Get the Roles to Add
                    foreach (var defaultRole in SecRoles)
                    {
                        Boolean matchToIgnoreAdd = false;
                        foreach (Entity role in currentRoleResults)
                        {
                            if (matchToIgnoreAdd)
                                break;

                            Guid roleGuid = Guid.Empty;
                            if (role != null && role.Attributes != null && role.Attributes["roleid"] != null)
                                roleGuid = (Guid)(role.Attributes["roleid"]);
                            var roleRecord = srv.RoleSet.FirstOrDefault(r => r.Id == roleGuid);
                            if ((roleRecord != null) && (roleRecord.Name != null))
                                matchToIgnoreAdd = (defaultRole == roleRecord.Name) ? true : false;
                        }
                        if (!matchToIgnoreAdd) //No Match, so add
                        {
                            var newRole = srv.RoleSet.FirstOrDefault(sr => sr.BusinessUnitId.Id == businessUnitId && sr.Name == defaultRole.ToString());
                            RolesToAssociate.Add(new EntityReference(Role.EntityLogicalName, newRole.Id));
                            addRoles++;
                            addedRoles = true;
                        }
                    }

                    //Get the Roles to Remove
                    foreach (Entity role in currentRoleResults)
                    {
                        Guid roleGuid = Guid.Empty;
                        if (role != null && role.Attributes != null && role.Attributes["roleid"] != null)
                            roleGuid = (Guid)(role.Attributes["roleid"]);
                        var roleRecord = srv.RoleSet.FirstOrDefault(r => r.Id == roleGuid);
                        Boolean matchToRemove = SecRoles.FirstOrDefault(r => r == roleRecord.Name) == null;
                        if (matchToRemove)
                        {
                            RolesToDisassociate.Add(new EntityReference(Role.EntityLogicalName, roleGuid));
                            removeRoles++;
                            removedRoles = true;
                        }
                    }

                    if (addedRoles)
                    {
                        OrganizationService.Associate(EntityLogicalName, thisId, new Relationship(EntityLogicalName.ToLower() + "roles_association"), RolesToAssociate);
                        Logger.WriteDebugMessage(String.Format("{0}/{1} {2}{3}", addRoles, totalRoles, displayMessage, thisRecordName));
                    }

                    if (removedRoles)
                    {
                        OrganizationService.Disassociate(EntityLogicalName, thisId, new Relationship(EntityLogicalName.ToLower() + "roles_association"), RolesToDisassociate);
                        Logger.WriteDebugMessage(String.Format("{0} role(s) removed from {1}", removeRoles, thisRecordName));
                    }
                }
                catch (FaultException<OrganizationServiceFault> ex)
                {
                    Logger.WriteToFile(ex.Message);
                }
                catch (Exception ex)
                {
                    Logger.WriteToFile(ex.Message);
                }
            }
        }
    
        #endregion

        #region Currently Unused Functions
        //For Create plugins there is no PRE image.
        //for anything else PRE images can exist on a pre-stage plugin but POST image does not.
        // For any other POST stage plugins PRE and POST exist.

        //since target only contains what was changed - the purpose of the helper is to be able to get a value of a field that may or may not have been updated.
        //For pre-stage plugins this can be difficult and require lots of coding as you want to use the target if the user updated the field and the property exists in the target - but if it
        //doesn't exist then the pre-image would hold the value.  That can be 4 lines of code, or more, for each attribute.

        //however for a post stage plugin if you want the post value then make the thisEntity and preEntity both contain the post image.
        //though not common it is possible for a post image to be different than the target.  
        //One scenario is phone number where maybe the phone number in target is 123-4567890 and the post it is 
        //(123) 456-7890.  This happens if a plugin changes the value prior to it getting to the db.

        //Use small logo
        public static string GetTSSLogo(IOrganizationService OrganizationService)
        {
            try
            {
                //create request to retrieve Webresource 
                QueryByAttribute requestWebResource = new QueryByAttribute
                {
                    EntityName = WebResource.EntityLogicalName,
                    ColumnSet = new ColumnSet(true),
                };
                requestWebResource.Attributes.AddRange("name");
                requestWebResource.Values.AddRange("cvt_TSS_Logo_Email");

                WebResource webResource = null;
                EntityCollection webResourceCollection = OrganizationService.RetrieveMultiple(requestWebResource);
                if (webResourceCollection.Entities.Count == 0)
                    throw new InvalidPluginExecutionException("Specified Webresource does not exist");

                webResource = (WebResource)webResourceCollection.Entities[0];
                return "<img src ='data:image/png;base64, " + webResource.Content + "' alt='TSS Logo' width='235' height='141'><br/>";
            }
            catch
            {
                return "";
            }
        }     
        #endregion
    }
}