﻿using System.Linq;
using mcsScheduling;
using Microsoft.Xrm.Sdk;
using System;
using System.ServiceModel;
using System.Collections.Generic;

namespace CVTPlugins
{
    public class CvtGroupPatientCreatePostStageRunner : MCSPlugins.PluginRunner
    {
        #region Constructor
        public CvtGroupPatientCreatePostStageRunner(IServiceProvider serviceProvider)
            : base(serviceProvider)
        {
        }
        #endregion

        #region Internal Methods/Properties
        internal void Execute()
        {
            try
            {                
               // if (PluginExecutionContext.Depth > 1 ){ return;}
                TracingService.Trace("at the top");
                Logger.setMethod = "Execute";
                //start the timing for the plugin
                Logger.WriteTxnTimingMessage(String.Format("Starting : {0}", GetType()));
                Logger.WriteDebugMessage("Starting");
                Logger.WriteDebugMessage(String.Format("At the top of: {0}", GetType()));

                //checks to make sure the registration is correct
                if (!PluginExecutionContext.InputParameters.Contains("Target") ||
                    !(PluginExecutionContext.InputParameters["Target"] is Entity))
                    return;

                var entity = (Entity)PluginExecutionContext.InputParameters["Target"];
                
                if (entity.LogicalName != cvt_grouppatient.EntityLogicalName)
                    throw new Exception("Target entity is not of type GroupPatient");

                var thisPatient = entity.ToEntity<cvt_grouppatient>();             
                var relatedSAId = McsHelper.getEntRefID("cvt_relatedserviceactivityid");
                var relatedPatSiteRoomId = McsHelper.getEntRefID("cvt_relatedpatientsiteroomid");
                Logger.WriteDebugMessage("SA: " + relatedSAId + " Room :" + relatedPatSiteRoomId);
              
                if ((relatedSAId != null) && (relatedPatSiteRoomId != null))
                {
                    Logger.WriteDebugMessage("Getting the Service Activity and PatSiteRoom IDs");                  
                    //start the timing for the submethod
                    Logger.WriteGranularTimingMessage("Starting UpdateCapacity");                   
                    //Create the Patient Site Rooms associated with this Service Activity
                    UpdateCapacity(thisPatient, relatedSAId, relatedPatSiteRoomId);
                    Logger.WriteGranularTimingMessage("Ending UpdateCapacity");
                }
                
                Logger.WriteDebugMessage("Ending");
                //End the timing for the plugin
                Logger.WriteTxnTimingMessage(String.Format("Ending : {0}", GetType()));
            }
            catch (FaultException<OrganizationServiceFault> ex)
            {
                Logger.WriteToFile(ex.Message);
                throw new InvalidPluginExecutionException(McsSettings.getUnexpectedErrorMessage);
            }
            catch (Exception ex)
            {
                if (ex.Message.StartsWith("custom"))
                {
                    Logger.WriteDebugMessage(ex.Message.Substring(6));
                    throw new InvalidPluginExecutionException(ex.Message.Substring(6));
                }
                else
                {
                    Logger.setMethod = "Execute";
                    Logger.WriteToFile(ex.Message);
                    throw new InvalidPluginExecutionException(ex.Message);
                }
            }
        }

        internal void UpdateCapacity(cvt_grouppatient thisPatient, Guid relatedSAId, Guid relatedPatSiteRoomId)
        {
            try
            {
                using (var srv = new Xrm(OrganizationService))
                {               
                    // Look up the Pat Site Room to be updated
                    Logger.setMethod = "UpdateCapacity";
                    Logger.WriteDebugMessage("Loading the PatSiteRoom");                 
                    var PatSiteRoom = srv.CreateQuery<mcs_patientsiteroom>().FirstOrDefault(c =>
                            c.Id == thisPatient.cvt_relatedpatientsiteroomid.Id);
                    if (PatSiteRoom == null)
                        return;
                   
                    //Look up that Patients related to the Patient Site Room
                    Logger.WriteDebugMessage("Creating collection of Patients.");
                    var PatSitePatients = from psp in srv.cvt_grouppatientSet
                                       where ((psp.cvt_relatedpatientsiteroomid.Id == relatedPatSiteRoomId)
                                              )
                                       select psp;

                    foreach (var psp in PatSitePatients)
                    {
                        Logger.WriteDebugMessage("Pat Site Patient: " + psp.cvt_name);
                    }  

                    //If there are no related Patients, set Patient count to 0, otherwise loop through and count the related Patients. 
                    var patientCount = 0;
                    if (PatSitePatients == null)
                    {
                        patientCount += 0;
                    }
                    else
                    {
                        foreach (var psp in PatSitePatients)
                        {
                            patientCount += 1;
                        }
                    }

                    var Capacity = PatSiteRoom.cvt_capacity;
                    var remainingCapacity = Capacity - patientCount;

                    var updatePatSiteRoom = new mcs_patientsiteroom()
                    {
                        Id = relatedPatSiteRoomId,
                        cvt_patients = patientCount,
                        cvt_remainingcapacity = remainingCapacity
                    };
                    OrganizationService.Update(updatePatSiteRoom);
                    
                    var RelatedServiceActivity = srv.CreateQuery<ServiceAppointment>().FirstOrDefault(c =>
                            c.Id == thisPatient.cvt_relatedserviceactivityid.Id); 

                    var ServiceActivityPatients = from sap in srv.cvt_grouppatientSet
                                                  where ((sap.cvt_relatedserviceactivityid.Id == RelatedServiceActivity.Id))
                                                  select sap;
                    foreach (var sap in ServiceActivityPatients)
                    {
                        Logger.WriteDebugMessage("Service Activity Patient: " + sap.cvt_name);
                    }  
                    var SApatientCount = 0;
                    if (ServiceActivityPatients == null)
                    {
                        SApatientCount += 0;
                    }
                    else
                    {
                        foreach (var p in ServiceActivityPatients)
                        {
                            SApatientCount += 1;
                        }
                    }

                    var currentSACapacity = RelatedServiceActivity.mcs_capacity;
                    var remainingSACapacity = currentSACapacity - SApatientCount;

                    var updateServiceActivity = new ServiceAppointment()
                    {
                        Id = relatedSAId,
                        cvt_patientcount = SApatientCount,
                        mcs_remainingcapacity = remainingSACapacity
                    };
                    
                    //Finally: Update the 'customers' field on the related Service Activity so we can accurately report on which Service Activity each Patient is realated to. 
                    //Creating a new Activity Party List and pulling all Group Patient / Patients and adding to the customer field on the Service Activity.                                       
                    List<ActivityParty> Patients = new List<ActivityParty>();
                    if (SApatientCount > 0)
                    {
                        foreach (var p in ServiceActivityPatients)
                        {
                            ActivityParty party = new ActivityParty();
                            EntityReference PatientReference = p.cvt_relatedpatientid;
                            party.PartyId = PatientReference;
                            Patients.Add(party);
                        }
                        if (Patients.Count > 0)
                        {
                            ActivityParty[] newPatientList = new ActivityParty[Patients.Count];
                            for (int i = 0; i < Patients.Count; i++)
                            {
                                newPatientList[i] = (ActivityParty)Patients[i];
                            }
                            updateServiceActivity.Customers = newPatientList;
                        }
                    }
                    //Update the Service Activity with new Capacity values and add patients to the customers field.                     
                    OrganizationService.Update(updateServiceActivity);
                    srv.SaveChanges();
                    Logger.WriteDebugMessage("Updated the Service Activity Patient Count.");                                    
                }
            }
            catch (FaultException<OrganizationServiceFault> ex)
            {
                Logger.WriteToFile(ex.Message);
            }
            catch (Exception ex)
            {
                Logger.WriteToFile(ex.Message);
            }
        }
        #endregion

        #region Additional Interface Methods/Properties
        /// <summary>
        /// Gets the Primary Entity which is triggering this plugin - not needed for this plugin except to satisfy the Abstract Class
        /// </summary>
        /// <returns>Entity Object which triggered plugin</returns>
        public override string McsSettingsDebugField
        {
            get { return "cvt_grouppatientplugin"; }
        }

        /// <summary>
        /// Gets the Entity pre-image - not needed for this plugin except to satisfy the Abstract Class
        /// </summary>
        /// <returns></returns>
        public override Entity GetPrimaryEntity()
        {
            return (Entity)PluginExecutionContext.InputParameters["Target"];
        }

        /// <summary>
        /// Used for Debugging - turns on or off creation of log records for this particular entity
        /// </summary>
        public override Entity GetSecondaryEntity()
        {
            return (Entity)PluginExecutionContext.InputParameters["Target"];
        }
        #endregion
    }
}