﻿using System.Linq;
using mcsScheduling;
using Microsoft.Xrm.Sdk;
using System;
using System.ServiceModel;
using Microsoft.Crm.Sdk.Messages;
using System.Collections.Generic;

namespace CVTPlugins
{
    public class McsServicesCreatePostStageRunner : MCSPlugins.PluginRunner
    {
        #region Constructor
        public McsServicesCreatePostStageRunner(IServiceProvider serviceProvider)
            : base(serviceProvider)
        {
        }
        #endregion

        #region Internal Methods/Properties
        internal void Execute()
        {
            try
            {
                TracingService.Trace("at the top");

                Logger.setMethod = "Execute";
                Logger.WriteTxnTimingMessage("Starting");

                Logger.WriteDebugMessage(String.Format("At the top of: {0}", GetType()));

                if (!PluginExecutionContext.InputParameters.Contains("Target") ||
                    !(PluginExecutionContext.InputParameters["Target"] is Entity))
                    return;

                var entity = (Entity)PluginExecutionContext.InputParameters["Target"];

                if (entity.LogicalName != mcs_services.EntityLogicalName)
                    throw new Exception("Target entity is not of type mcs_services");

                var tsa = entity.ToEntity<mcs_services>();
                if (tsa != null && tsa.cvt_RelatedMasterId == null)
                {
                    Logger.WriteDebugMessage("not doing quick create");
                    return;
                }

                Logger.WriteDebugMessage("Doing quick create");
                Logger.WriteGranularTimingMessage("Starting QuickCreateMsta");
                CreateMasterProviderResourceGroups(tsa);
                Logger.WriteGranularTimingMessage("Ending QuickCreateMsta");
                
                //Run enable InterVISN TSA - the method itself will exit execution if criteria are not met
                EnableInterVISNTSA(tsa);

                Logger.setMethod = "Execute";

                Logger.WriteTxnTimingMessage("Ending");

            }
            catch (FaultException<OrganizationServiceFault> ex)
            {
                Logger.WriteToFile(ex.Message);
                throw new InvalidPluginExecutionException(ex.Message);

            }
            catch (Exception ex)
            {
                Logger.WriteToFile(ex.Message);
                throw new InvalidPluginExecutionException(ex.Message);
            }
        }

        internal void EnableInterVISNTSA(mcs_services tsa)
        {
            Logger.setMethod = "Enable InterVISN TSA";
            Logger.WriteDebugMessage("Starting InterVISN TSA");
            using (var srv = new Xrm(OrganizationService))
            {
                var proFacility = srv.mcs_facilitySet.FirstOrDefault(f => f.Id == tsa.cvt_ProviderFacility.Id);
                var patFacility = tsa.cvt_PatientFacility != null ? srv.mcs_facilitySet.FirstOrDefault(f => f.Id == tsa.cvt_PatientFacility.Id) : null;
                if ( patFacility == null || proFacility.mcs_BusinessUnitId.Id == patFacility.mcs_BusinessUnitId.Id) //Only share if the VISN of the 2 facilities dont match
                    return;
                //TO-DO: WHICH TEAM DO WE SHARE WITH AND WHAT EXACTLY DO WE SHARE - READ, WRITE, APPEND?
                var FacilityTeams = srv.TeamSet.Where(t => t.cvt_Facility.Id == patFacility.Id && 
                    (t.cvt_ServiceType == null || t.cvt_ServiceType.Id == tsa.cvt_servicetype.Id)).ToList();
                AccessRights approverRights = AccessRights.AppendAccess | AccessRights.AppendToAccess | AccessRights.WriteAccess | AccessRights.ReadAccess;

                //set the following teams so they can be shared with
                var patFTCTeam = new Team();
                var patCPTeam = new Team();
                var patSCTeam = new Team();
                var patCoSTeam = new Team();
                var patSchedTeam = new Team();

                foreach (var team in FacilityTeams)
                {
                    if (team.cvt_Type != null && team.cvt_Type.Value == 917290000)
                        patFTCTeam = team;
                    if (team.cvt_Type != null && team.cvt_Type.Value == 917290001)
                        patCPTeam = team;
                    if (team.cvt_Type != null && team.cvt_Type.Value == 917290002)
                        patSCTeam = team;
                    if (team.cvt_Type != null && team.cvt_Type.Value == 917290003)
                        patCoSTeam = team;
                    if (team.cvt_Type != null && team.cvt_Type.Value == 917290005)
                        patSchedTeam = team;

                }

                //Give Access to all of the proper teams
                GrantAccess(patFTCTeam, tsa, approverRights);
                GrantAccess(patSCTeam, tsa, approverRights);
                GrantAccess(patCPTeam, tsa, approverRights);
                GrantAccess(patCoSTeam, tsa, approverRights);
                GrantAccess(patSchedTeam, tsa, AccessRights.AppendToAccess | AccessRights.AppendAccess | AccessRights.ReadAccess);

            }
            
        }

        /// <summary>
        /// Shares the privilege specified in "accessType" for the record in "Target" with the team in "Team"
        /// </summary>
        /// <param name="team">Team who will be shared with</param>
        /// <param name="target">Record that will be shared</param>
        /// <param name="accessType">privilege that will be shared</param>
        internal void GrantAccess(Team team, mcs_services target, AccessRights accessType)
        {
            if (team == null || team.Id == new Guid())
            {
                Logger.WriteDebugMessage("Team not found");
                return;
            }
            Logger.setMethod = "Grant Access " + accessType.ToString();
            Logger.WriteDebugMessage(String.Format("Granting {0} privileges to {1}", accessType.ToString(), team.Name));
            // Grant the Team read access to the created TSA.
            var grantAccessRequest = new GrantAccessRequest
            {
                PrincipalAccess = new PrincipalAccess
                {
                    AccessMask = accessType,
                    Principal = new EntityReference()
                    {
                        Id = team.Id,
                        Name = team.Name,
                        LogicalName = team.LogicalName
                    }
                },
                Target = new EntityReference()
                {
                    Id = target.Id,
                    Name = target.mcs_name,
                    LogicalName = target.LogicalName
                }
            };

            OrganizationService.Execute(grantAccessRequest);
            Logger.WriteDebugMessage("Privilege Granted");
        }

        internal static cvt_providerresourcegroup GetNewProviderResourceGroup(mcs_services tsa, 
            cvt_providerresourcegroup providerResourceGroup)
        {
            
                var newProviderResourceGroup = new cvt_providerresourcegroup()
                    {
                        cvt_name = providerResourceGroup.cvt_name,
                        cvt_RelatedResourceGroupid = providerResourceGroup.cvt_RelatedResourceGroupid,
                        cvt_RelatedResourceId = providerResourceGroup.cvt_RelatedResourceId,
                        cvt_relatedsiteid = providerResourceGroup.cvt_relatedsiteid,
                        cvt_RelatedTSAid = new EntityReference(mcs_services.EntityLogicalName, tsa.Id),
                        cvt_RelatedUserId = providerResourceGroup.cvt_RelatedUserId,
                        cvt_Type = providerResourceGroup.cvt_Type,
                        cvt_TSAResourceType = providerResourceGroup.cvt_TSAResourceType,
                        cvt_constraintgroupguid = providerResourceGroup.cvt_constraintgroupguid,
                        cvt_resourcespecguid = providerResourceGroup.cvt_resourcespecguid
                    };


                return newProviderResourceGroup;
            
        }

        internal void CreateMasterProviderResourceGroups(mcs_services tsa)
        {
            try
            {
                Logger.setMethod = "QuickCreatePatientGroupResource";
                using (var context = new Xrm(OrganizationService))
                {
                    var masterTsa = context.CreateQuery<cvt_mastertsa>().FirstOrDefault(c => 
                        c.Id == tsa.cvt_RelatedMasterId.Id);

                    if (masterTsa == null)
                        return;

                    context.LoadProperty(masterTsa, "cvt_cvt_mastertsa_cvt_providerresourcegroup_RelatedTSAId");

                    if (masterTsa.cvt_cvt_mastertsa_cvt_providerresourcegroup_RelatedTSAId == null || 
                        !masterTsa.cvt_cvt_mastertsa_cvt_providerresourcegroup_RelatedTSAId.Any())
                        return;

                    foreach (var newProviderResourceGroup in
                       masterTsa.cvt_cvt_mastertsa_cvt_providerresourcegroup_RelatedTSAId.Select(providerResourceGroup =>
                            GetNewProviderResourceGroup(tsa, providerResourceGroup)))
                    {
                        Logger.WriteDebugMessage("Adding Resource");
                        context.AddObject(newProviderResourceGroup);
                    }
                    Logger.WriteDebugMessage("Saving Changes");

                    context.SaveChanges();
                }
            }
            catch (FaultException<OrganizationServiceFault> ex)
            {
                Logger.WriteToFile(ex.Message);
                throw new InvalidPluginExecutionException(ex.Message);

            }
            catch (Exception ex)
            {
                Logger.WriteToFile(ex.Message);
                throw new InvalidPluginExecutionException(ex.Message);

            }
        }

        public override string McsSettingsDebugField
        {
            get { return "mcs_tsaplugin"; }
        }

        public override Entity GetPrimaryEntity()
        {
            return (Entity)PluginExecutionContext.InputParameters["Target"];
        }

        public override Entity GetSecondaryEntity()
        {
            return (Entity)PluginExecutionContext.InputParameters["Target"];
        }
        #endregion
    }
}
