﻿using System;
using System.Linq;
using System.ServiceModel;
using Microsoft.Xrm.Sdk;
using MCSPlugins;
using mcsScheduling;
using Microsoft.Xrm.Sdk.Query;
using System.Collections.Generic;
using Microsoft.Crm.Sdk.Messages;
using Microsoft.Xrm.Sdk.Messages;
using MCSShared;

namespace CVTPlugins
{
    class CvtServiceActivityCreatePostStageRunner : PluginRunner
    {
        #region Constructor
        public CvtServiceActivityCreatePostStageRunner(IServiceProvider serviceProvider)
            : base(serviceProvider)
        {}

        #endregion

        #region Internal Methods/Properties
        /// <summary>
        /// This method determines whether or not to run the Generate Appointments and Clear Patient Resources on the Service Activity being created based on if it is a group appointment or not.  If the Service Activity is generated by any automated process, this function will return immediately before executing logic.
        /// </summary>
        internal void Execute()
        {
            try
            {
                //if (PluginExecutionContext.Depth > 1) { return; }

                TracingService.Trace("at the top");

                Logger.setMethod = "Execute";
                //start the timing for the plugin
                Logger.WriteTxnTimingMessage(String.Format("Starting : {0}", GetType()));
                Logger.WriteDebugMessage("Starting");

                Logger.WriteDebugMessage(String.Format("At the top of: {0}", GetType()));

                //checks to make sure the registration is correct
                if (!PluginExecutionContext.InputParameters.Contains("Target") ||
                    !(PluginExecutionContext.InputParameters["Target"] is Entity))
                    return;

                var entity = (Entity)PluginExecutionContext.InputParameters["Target"];

                if (entity.LogicalName != ServiceAppointment.EntityLogicalName)
                    throw new Exception("Target entity is not of type ServiceAppointment");

                Logger.WriteDebugMessage("Checking if this is a Group Appt");
                var thisServiceActivity = entity.ToEntity<ServiceAppointment>();
                var groupAppt = thisServiceActivity.mcs_groupappointment.Value;

                if (groupAppt == true)
                {
                    Logger.WriteGranularTimingMessage("Starting Appointment Generation");

                    //Generate 1 appointment per patient site resource group, save the list to be used as the resources to remove from the service activity (aka blocked)
                    var blocked = SetupAppointments(thisServiceActivity);
                    Logger.WriteGranularTimingMessage("Ending Appointment Generation");
                    //Now that the appointments are set up, go back to the service activity and remove all patient site resources
                    ClearPatientSideResources(thisServiceActivity, blocked);
                    Logger.WriteGranularTimingMessage("Ending Clearing Patient Site Resources");
                }
                Logger.WriteDebugMessage("Ending Create Plugin");
            }
            catch (FaultException<OrganizationServiceFault> ex)
            {
                Logger.WriteToFile(ex.Message);
                throw new InvalidPluginExecutionException(McsSettings.getUnexpectedErrorMessage);
            }
            catch (Exception ex)
            {
                if (ex.Message.StartsWith("custom"))
                {
                    Logger.WriteDebugMessage(ex.Message.Substring(6));
                    throw new InvalidPluginExecutionException(ex.Message.Substring(6));
                }
                else
                {
                    Logger.setMethod = "Execute";
                    Logger.WriteToFile(ex.Message);
                    throw new InvalidPluginExecutionException(ex.Message);
                }
            }
        }

        /// <summary>
        /// Read TSA listed on Service Activity, get the All Required groups, then set up an appointment for each one
        /// </summary>
        /// <param name="SA"></param>
        /// <returns>List of activity parties that correspond to the patient site resources</returns>
        /// <remarks>return list used later as the list of resources to be removed from the service activity "Resources" field, so only provider resources are booked on the service activity - patient site resources are booked on the child appointment records generated in this method</remarks>
        internal List<ActivityParty> SetupAppointments(ServiceAppointment SA)
        {
            var resourcesForRemoval = new List<ActivityParty>();
            using (var srv = new Xrm(OrganizationService))
            {
                var pats = srv.cvt_patientresourcegroupSet.Where(pat => pat.cvt_RelatedTSAid.Id == SA.mcs_relatedtsa.Id);
                foreach (var pat in pats)
                {
                    if (pat.cvt_RelatedResourceGroupid != null)
                    {
                        var ResourceGroup = srv.mcs_resourcegroupSet.FirstOrDefault(gr => gr.Id == pat.cvt_RelatedResourceGroupid.Id);
                        //This method both generates the appointment for the patient side resource, 
                        //as well as adds the resource to the list of resources to be removed from the Service Activity
                        if (ResourceGroup != null)
                            resourcesForRemoval.AddRange(GenerateGroupAppointment(ResourceGroup, SA));
                    }
                    //Here is where we add individual resources to list of ones to be removed from Service Activity
                    var participant = new Entity();
                    if (pat.cvt_RelatedResourceId != null)
                        participant = CvtHelper.getPartyParticipantFromResource(pat.cvt_RelatedResourceId.Id, srv);
                    if (pat.cvt_RelatedUserId != null)
                        participant = CvtHelper.getPartyParticipantFromResource(pat.cvt_RelatedUserId.Id, srv);
                    if (participant.LogicalName != null)
                    {
                        var party = new ActivityParty()
                        {
                            PartyId = new EntityReference(participant.LogicalName, participant.Id)
                        };
                        resourcesForRemoval.Add(party);
                    }
                }
            }
            return resourcesForRemoval;
        }

        /// <summary>
        /// accepts the Service Activity and 1 group as parameters.  Called 1 time per patient side resource group.  Creates an Appointment record for the group.
        /// </summary>
        /// <param name="group">The resource group (Patient All Required Group from TSA) that will be parsed and added to the Appointment</param>
        /// <param name="SA">Service Activity that will be the parent of the Appointment</param>
        /// <returns>Returns the list of ActivityParty records that got booked in the "Required" field of the appointment records</returns>
        /// <remarks>Appointment has to be saved in Open:Requested Status, and then State Change is initiated to get it to Scheduled Status</remarks>
        internal List<ActivityParty> GenerateGroupAppointment(mcs_resourcegroup group, ServiceAppointment SA)
        {
            var participants = new List<ActivityParty>();
            using (var context = new Xrm(OrganizationService))
            {
                var site = group.mcs_relatedSiteId;
                var GroupResources = context.mcs_groupresourceSet.Where(gr => gr.mcs_relatedResourceGroupId.Id == group.Id);
                foreach (var gr in GroupResources)
                {
                    var participant = new Entity();
                    if (gr.mcs_RelatedResourceId != null)
                        participant = CvtHelper.getPartyParticipantFromResource(gr.mcs_RelatedResourceId.Id, context);
                    else if (gr.mcs_RelatedUserId != null)
                        participant = CvtHelper.getPartyParticipantFromResource(gr.mcs_RelatedUserId.Id, context);
                    var party = new ActivityParty()
                    {
                        PartyId = new EntityReference(participant.LogicalName, participant.Id),
                        ParticipationTypeMask = new OptionSetValue(5)
                    };
                    participants.Add(party);
                }
                var appt = new Appointment(){
                    ScheduledStart = SA.ScheduledStart,
                    ScheduledEnd = SA.ScheduledEnd,
                    ScheduledDurationMinutes = SA.ScheduledDurationMinutes,
                    cvt_serviceactivityid = new EntityReference(ServiceAppointment.EntityLogicalName, SA.Id),
                    RequiredAttendees = participants,
                    cvt_ResourceGroup = new EntityReference(mcs_resourcegroup.EntityLogicalName, group.Id),
                    Subject = SA.Subject + " for " + group.mcs_name,
                    cvt_Site = site
                };
                var apptId = OrganizationService.Create(appt);
                //You have to create the appointment with open status and then call SetState to mark it as booked
                SetStateRequest ssr = new SetStateRequest()
                {
                    State = new OptionSetValue(3),
                    Status = new OptionSetValue(5),
                    EntityMoniker = new EntityReference(Appointment.EntityLogicalName, apptId) 
                };
                OrganizationService.Execute(ssr);
            }
            return participants;
        }

        /// <summary>
        /// This method removes the resources from the "Resources" field on the service activity based on the "blocked" list and resaves it
        /// </summary>
        /// <param name="SA">The service Activity that is being acted on</param>
        /// <param name="blocked">The list of resources that are Patient Side resources (and thus have already been booked into appointments) which gets removed from the "Resources" field, leaving only the provider resources on the service activity</param>
        /// <remarks>Have to Set PluginExecution Object to empty list to clear it before using new object with correct list otherwise it appends new list to old list</remarks>
        internal void ClearPatientSideResources(ServiceAppointment SA, List<ActivityParty> blocked)
        {
            var resList = new List<ActivityParty>();
            var resources = (EntityCollection)SA.Attributes["resources"];

            //use idList to compare Party Ids in entity collection to PartyIds from partyList (since the list items on the appointments and resources on the Service Appointment will always be different activity party items)
            var idList = new List<Guid>();
            foreach (var item in blocked)
                idList.Add(item.PartyId.Id);
            foreach (var res in resources.Entities)
            {
                var resId = ((EntityReference)(res.Attributes["partyid"])).Id;
                if (!idList.Contains(resId))
                {
                    var participant = new Entity();
                    using (var srv = new Xrm(OrganizationService))
                    {
                        participant = srv.EquipmentSet.FirstOrDefault(e => e.Id == resId);
                        if (participant == null)
                            participant = srv.SystemUserSet.FirstOrDefault(u => u.Id == resId);
                    }
                    var party = new ActivityParty()
                    {
                        PartyId = new EntityReference(participant.LogicalName, participant.Id),
                        ParticipationTypeMask = new OptionSetValue(10)
                    };
                    resList.Add(party);
                }
            }

            //Have to Set PluginExecution Object to empty list to clear it before using new object with correct list otherwise it appends new list to old list
            SA.Resources = new List<ActivityParty>();
            OrganizationService.Update(SA);
            ServiceAppointment ServAppt = new ServiceAppointment()
            {
                Id = SA.Id,
                Resources = resList
            };
            OrganizationService.Update(ServAppt);
        }

        #endregion

        #region Additional Interface Methods/Properties
        public override Entity GetPrimaryEntity()
        {
            return (Entity)PluginExecutionContext.InputParameters["Target"];
        }

        public override Entity GetSecondaryEntity()
        {
            return (Entity)PluginExecutionContext.InputParameters["post"];
        }

        public override string McsSettingsDebugField
        {
            get { return "cvt_serviceactivityplugin"; }
        }
        #endregion
    }
}