﻿using System;
using System.ServiceModel;
using Microsoft.Xrm.Sdk;
using mcsScheduling;
using System.Linq;

namespace MCSPlugins
{
    public class McsGroupResourceDeletePreStageRunner : PluginRunner
    {
        #region Constructor
        public McsGroupResourceDeletePreStageRunner(IServiceProvider serviceProvider)
            : base(serviceProvider)
        {
        }       
        #endregion
        
        #region Internal Methods/Properties
        internal void Execute()
        {
            try
            {
                Logger.setMethod = "Execute";

                Logger.WriteTxnTimingMessage("Starting");

                if (PluginExecutionContext.Depth > 1) { return; }

                if (PluginExecutionContext.InputParameters.Contains("Target") &&
                    PluginExecutionContext.InputParameters["Target"] is EntityReference)
                {

                    Logger.WriteDebugMessage(String.Format("At the top of:{0}", GetType()));
                    Logger.WriteGranularTimingMessage("Starting UpdateResourceGroup");
                    UpdateResourceGroup(PluginExecutionContext.PrimaryEntityId);
                    Logger.WriteGranularTimingMessage("Ending UpdateResourceGroup");
                }

                Logger.setMethod = "Execute";

                Logger.WriteTxnTimingMessage("Ending");

            }
            catch (FaultException<OrganizationServiceFault> ex)
            {
                Logger.WriteToFile(ex.Message);
                throw new InvalidPluginExecutionException(McsSettings.getUnexpectedErrorMessage + ":" + ex.Message);
            }
            catch (Exception ex)
            {
                if (ex.Message.StartsWith("custom"))
                {
                    Logger.WriteDebugMessage(ex.Message.Substring(6));

                    throw new InvalidPluginExecutionException(ex.Message.Substring(6));
                }
                else
                {
                    Logger.setMethod = "Execute";
                    Logger.WriteToFile(ex.Message);
                    throw new InvalidPluginExecutionException(McsSettings.getUnexpectedErrorMessage + ":" + ex.Message);
                }
            }
        }

        internal void UpdateResourceGroup(Guid thisId)
        {
            Logger.WriteDebugMessage("starting UpdateResourceGroup");

            using (var srv = new Xrm(OrganizationService))
            {
                                   
                    var thisGroupResource = srv.mcs_groupresourceSet.FirstOrDefault(i => i.Id == thisId);
                    //First check to see if Related Resource Group is related to Pat/Prov Site Resources. 
                    var RelatedResourceGroup = srv.mcs_resourcegroupSet.FirstOrDefault(i => i.Id == thisGroupResource.mcs_relatedResourceGroupId.Id);

                    //Checking to see if any Patient Site Resources have this target MCS Group Resource / Resource Group associated with it. 
                    var patResource = srv.cvt_patientresourcegroupSet.FirstOrDefault(i => i.cvt_RelatedResourceGroupid.Id == RelatedResourceGroup.Id);
                    //If an associated Patient Site Resource does exist, we will throw an exception with a message and stop the delete of the MCS Resource, to prevent orphan data. 
                    if (patResource != null)
                    {
                        Logger.WriteDebugMessage("Patient Resource Exists:" + patResource.cvt_name);
                        throw new InvalidPluginExecutionException("customPlease check Left Nav on related Resource Group form for Patient Site Resources that the related Resource Group is associated with. Resource cannot be deleted until associations are removed.");
                    }
                    //Checking to see if any Provider Site Resources have this target MCS Gropu Resource / Resource Group associated with it. 
                    var proResource = srv.cvt_providerresourcegroupSet.FirstOrDefault(i => i.cvt_RelatedResourceGroupid.Id == RelatedResourceGroup.Id);
                    //If an associated Provider Site Resource does exist, we will throw an exception with a message and stop the delete of the MCS Resource, to prevent orphan data.
                    if (proResource != null)
                    {
                        Logger.WriteDebugMessage("Provider Resource Exists:" + proResource.cvt_name);
                        throw new InvalidPluginExecutionException("customPlease check Left Nav on related Resource Group form for Provider Site Resources that the related Resource Group is associated with. Resource cannot be deleted until associations are removed.");
                    }


                    Logger.setMethod = "Update System Group Resource";
                    var builder = new System.Text.StringBuilder("<Constraints>");
                    builder.Append("<Constraint>");
                    builder.Append("<Expression>");
                    var getResources = from resGroups in srv.mcs_groupresourceSet
                                       join mcsResourcs in srv.mcs_resourceSet on resGroups.mcs_RelatedResourceId.Id equals mcsResourcs.mcs_resourceId.Value
                                       where resGroups.mcs_relatedResourceGroupId.Id == RelatedResourceGroup.Id && resGroups.statecode == 0
                                       select new
                                       {
                                           mcsResourcs.mcs_name,
                                           mcsResourcs.mcs_relatedResourceId,
                                           resGroups.mcs_groupresourceId
                                       };

                    string resources = null;
                    var count = 0;
                    foreach (var mcsResourcs in getResources)
                    {
                        if (mcsResourcs.mcs_groupresourceId.Value != thisId)
                        {
                            //First add the resources to the text field for the view. 
                            resources += mcsResourcs.mcs_relatedResourceId.Name.ToString();
                            resources += " ; ";
                            
                            builder.Append(count > 0 ? " || resource[\"Id\"] == " : "<Body>resource[\"Id\"] == ");
                            builder.Append(mcsResourcs.mcs_relatedResourceId.Id.ToString("B"));
                            count += 1;
                        }
                    }

                    var getUsers = from resGroups in srv.mcs_groupresourceSet
                                   join Users in srv.SystemUserSet on resGroups.mcs_RelatedUserId.Id equals Users.Id
                                   where resGroups.mcs_relatedResourceGroupId.Id == thisId && resGroups.statecode == 0
                                   select new
                                   {
                                       //  Users.FirstName,
                                       //  Users.LastName,
                                       //  Users.Id,
                                       resGroups.mcs_RelatedUserId,
                                       resGroups.mcs_name,
                                       resGroups.mcs_groupresourceId

                                   };

                    foreach (var User in getUsers)
                    {
                        if (User.mcs_groupresourceId.Value != thisId)
                        {
                            resources += User.mcs_RelatedUserId.Name.ToString();
                            resources += " ; ";
                            
                            builder.Append(count > 0 ? " || resource[\"Id\"] == " : "<Body>resource[\"Id\"] == ");
                            builder.Append(User.mcs_RelatedUserId.Id.ToString("B"));
                            count += 1;
                        }

                    }

                    if (count == 0)
                    {
                        builder.Append("<Body>false");

                    }

                    builder.Append("</Body>");
                    builder.Append("<Parameters>");
                    builder.Append("<Parameter name=\"resource\" />");
                    builder.Append("</Parameters>");
                    builder.Append("</Expression>");
                    builder.Append("</Constraint>");
                    builder.Append("</Constraints>");

                    Logger.WriteDebugMessage("About to get mcs_resourcegroup");
                    var resourceGroup = srv.mcs_resourcegroupSet.FirstOrDefault(i => i.Id == RelatedResourceGroup.Id);
                    if (resourceGroup == null) return;
                    Logger.WriteDebugMessage("got mcs_resourcegroup");
                    if (resourceGroup.mcs_RelatedResourceGroupId != null)
                    {
                        var myRealResourceGroup = (EntityReference)resourceGroup["mcs_relatedresourcegroupid"];
                        var group = new ConstraintBasedGroup
                        {
                            Id = myRealResourceGroup.Id,
                            Constraints = builder.ToString()
                        };
                        var updateRG = new mcs_resourcegroup
                        {

                            Id = resourceGroup.Id,
                            cvt_resources = resources
                        };
                        
                        OrganizationService.Update(updateRG);
                        OrganizationService.Update(group);
                        //whereami = "updated";
                        Logger.WriteDebugMessage("System Resource Updated with " + count + " resources");

                    }

                
            }
        }
 
        public override string McsSettingsDebugField
        {
            get { return "mcs_groupresourceplugin"; }
        }

        public override Entity GetPrimaryEntity()
        {
            return (Entity)PluginExecutionContext.PreEntityImages["pre"];
        }

        public override Entity GetSecondaryEntity()
        {
            return (Entity)PluginExecutionContext.PreEntityImages["pre"];
        }
        #endregion
    }
}
