﻿using Microsoft.Crm.Sdk.Messages;
using Microsoft.Xrm.Sdk;
using System;
using System.ServiceModel;
using mcsScheduling;
using System.Linq;
using Microsoft.Xrm.Sdk.Query;
using MCSShared;

namespace MCSPlugins
{
    public class McsResourceCreatePostStageRunner : PluginRunner
    {
        #region Constructor
        public McsResourceCreatePostStageRunner(IServiceProvider serviceProvider)
            : base(serviceProvider)
        {
        }
        #endregion

        #region Methods
        internal void Execute()
        {
            //This plugin runs after the mcs_resource has been created.
            //
            // It accomplishes the following:
            //1.) If the user had hit create resource from the resource group, create a group resource record to join the resource and resource group.
            //2.) If the user had hit create resource from the provider resource group grid, create a provider resource group record to join the TSA or MTSA to the resource
            //      a.) There are 2 text fields which are used to designate if it was from a TSA or MTSA
            //3.) If the user had hit create resource from the patient resource group grid, create a patient resource group record to join the TSA or MTSA to the resource
            //4.) Create a system resource (really equipment) to actually have the calendar for this resource
            //5.) Find the default calendar and make the new system resource have that calendar instead of the OOB one.
            //6.) Assign the resource to the primary team of the user.
            try
            {
                Logger.setMethod = "Execute";
                Logger.WriteTxnTimingMessage("Starting");

                if (PluginExecutionContext.InputParameters.Contains("Target") && PluginExecutionContext.InputParameters["Target"] is Entity)
                {
                    Logger.WriteDebugMessage(String.Format("At the top of:{0}", GetType()));
                
                        //Create the system resource
                        Logger.WriteGranularTimingMessage("Starting CreateSysResource");
                        CreateSysResource(PluginExecutionContext.PrimaryEntityId);
                        Logger.WriteGranularTimingMessage("Ending CreateSysResource");
                                       
                    //see if they are creating this from resource group
                    var quickCreateOptionResourceGroup = McsHelper.getStringValue("mcs_resourcegroupguid");
                    var doQuickCreateResourceGroup = quickCreateOptionResourceGroup != null;
                    Logger.WriteDebugMessage(String.Format("doQuickCreate:{0}{1}", doQuickCreateResourceGroup, doQuickCreateResourceGroup));

                    if (doQuickCreateResourceGroup)
                    {
                        //if they are, create the group resource record
                        Logger.WriteDebugMessage("about to do resource group quick create");
                        Logger.WriteGranularTimingMessage("Starting quickCreateGroupResource");
                        QuickCreateGroupResource(PluginExecutionContext.PrimaryEntityId);
                        Logger.WriteGranularTimingMessage("Ending quickCreateGroupResource");
                    }
                    else
                        Logger.WriteDebugMessage("not doing resource group quick create");

                      //if the mcs_tsaguid is present then it is because they are doing a quick create
                    var quickCreateOptionPatProv = McsHelper.getStringValue("mcs_tsaguid");
                    //but we have to know which one, pat or prov.
                    //These 3 fields are populated from jscript on a ribbon button.
                    var quickCreatePat = McsHelper.getBoolValue("mcs_createpatientr");
                    var quickCreateProv = McsHelper.getBoolValue("mcs_createproviderr");                
                    var doQuickCreatePatProv = quickCreateOptionPatProv != null;

                    if (!doQuickCreatePatProv)
                    {
                        Logger.WriteDebugMessage("Not doing Quick Create for TSA, maybe MTSA");
                        //if the mcs_mastertsaguid is present then it is because they are doing a quick create
                        quickCreateOptionPatProv = McsHelper.getStringValue("cvt_mastertsaguid");

                        doQuickCreatePatProv = quickCreateOptionPatProv != null;
                    }

                    Logger.WriteDebugMessage(String.Format("doQuickCreatePatProv:{0}{1}", doQuickCreatePatProv, quickCreateOptionPatProv));

                    if (doQuickCreatePatProv)
                    {
                        if (quickCreatePat)
                        {
                            Logger.WriteDebugMessage("about to do quick create for Pat");
                            Logger.WriteGranularTimingMessage("Starting QuickCreatePatientGroupResource");

                            //call the method to do a quick create for pat group
                            QuickCreatePatientGroupResource(PluginExecutionContext.PrimaryEntityId);

                            Logger.WriteGranularTimingMessage("Ending QuickCreatePatientGroupResource");
                        }
                        else
                        {
                            Logger.WriteDebugMessage("quickCreateProv:" + quickCreateProv);
                            if (quickCreateProv)
                            {
                                Logger.WriteGranularTimingMessage("Starting QuickCreateProviderGroupResource");
                                Logger.WriteDebugMessage("about to do quick create for Prov");

                                //call the method to do a quick create for provider group
                                QuickCreateProviderGroupResource(PluginExecutionContext.PrimaryEntityId);

                                Logger.WriteGranularTimingMessage("Ending QuickCreateProviderGroupResource");
                            }
                        }
                    }
                }

                //This snippet is used to updated the Identifier field so that it is viewable 
                var updatedResource = new mcs_resource()
                {
                    Id = PluginExecutionContext.PrimaryEntityId,
                    cvt_Identifier = PluginExecutionContext.PrimaryEntityId.ToString()
                };
                OrganizationService.Update(updatedResource);
                Logger.setMethod = "Execute";
                Logger.WriteTxnTimingMessage("Ending");

            }
            catch (FaultException<OrganizationServiceFault> ex)
            {
                Logger.WriteToFile(ex.Message);
                throw new InvalidPluginExecutionException(ex.Message);
            }
            catch (Exception ex)
            {
                if (ex.Message.StartsWith("custom"))
                {
                    Logger.WriteDebugMessage(ex.Message.Substring(6));
                    throw new InvalidPluginExecutionException(ex.Message.Substring(6));
                }
                else
                {
                    Logger.setMethod = "Execute";
                    Logger.WriteToFile(ex.Message);
                    throw new InvalidPluginExecutionException(ex.Message);
                }
            }
        }

        internal void QuickCreateGroupResource(Guid primaryEntityId)
        {
            try
            {
                var newGroupResource = new Entity("mcs_groupresource");

                newGroupResource["mcs_relatedresourcegroupid"] = new EntityReference("mcs_resourcegroup", new Guid(McsHelper.getStringValue("mcs_resourcegroupguid")));
                newGroupResource["mcs_relatedsiteid"] = new EntityReference("mcs_site", McsHelper.getEntRefID("mcs_relatedsiteid"));
                newGroupResource["mcs_relatedresourceid"] = new EntityReference("mcs_resource", primaryEntityId);
                newGroupResource["mcs_name"] = McsHelper.getStringValue("mcs_name");
                newGroupResource["mcs_type"] = new OptionSetValue(McsHelper.getOptionSetValue("mcs_type"));
               // newGroupResource["mcs_timezone"] = McsHelper.getIntValue("mcs_timezone");
                OrganizationService.Create(newGroupResource);

                Logger.WriteDebugMessage("created group resource record");
            }
            catch (FaultException<OrganizationServiceFault> ex)
            {
                Logger.WriteToFile(ex.Message);
            }
            catch (Exception ex)
            {
                Logger.WriteToFile(ex.Message);
            }
        }

        internal void QuickCreatePatientGroupResource(Guid primaryEntityId)
        {
            try
            {
                Logger.setMethod = "QuickCreatePatientGroupResource";
         
                var newGroupResource = new Entity("cvt_patientresourcegroup");
                newGroupResource["cvt_tsaresourcetype"] = new OptionSetValue(0);
                newGroupResource["cvt_type"] = new OptionSetValue(McsHelper.getOptionSetValue("mcs_type"));
                newGroupResource["cvt_tsaresourcetype"] = new OptionSetValue(1);
                newGroupResource["cvt_type"] = new OptionSetValue(McsHelper.getOptionSetValue("mcs_type"));

                var tsapresent = McsHelper.getStringValue("mcs_tsaguid");
                if (tsapresent != null)
                    newGroupResource["cvt_relatedtsaid"] = new EntityReference("mcs_services", new Guid(tsapresent));

                newGroupResource["cvt_relatedsiteid"] = new EntityReference("mcs_site", McsHelper.getEntRefID("mcs_relatedsiteid"));
                newGroupResource["cvt_relatedresourceid"] = new EntityReference("mcs_resource", primaryEntityId);
                newGroupResource["cvt_name"] = McsHelper.getStringValue("mcs_name");

                OrganizationService.Create(newGroupResource);
                Logger.WriteDebugMessage("created patient resource group record");
            }
            catch (FaultException<OrganizationServiceFault> ex)
            {
                Logger.WriteToFile(ex.Message);
                throw new InvalidPluginExecutionException(ex.Message);
            }
            catch (Exception ex)
            {
                Logger.WriteToFile(ex.Message);
                throw new InvalidPluginExecutionException(ex.Message);
            }
        }

        internal void QuickCreateProviderGroupResource(Guid primaryEntityId)
        {
            try
            {
                Logger.setMethod = "QuickCreateProviderGroupResource";
                var newGroupResource = new Entity("cvt_providerresourcegroup");
                newGroupResource["cvt_tsaresourcetype"] = new OptionSetValue(1);
                newGroupResource["cvt_type"] = new OptionSetValue(McsHelper.getOptionSetValue("mcs_type"));

                var mastertsapresent = McsHelper.getStringValue("cvt_mastertsaguid");
                if (mastertsapresent != null)
                    newGroupResource["cvt_relatedmastertsaid"] = new EntityReference("cvt_mastertsa", new Guid(mastertsapresent));

                var tsapresent = McsHelper.getStringValue("mcs_tsaguid");
                if (tsapresent != null)
                    newGroupResource["cvt_relatedtsaid"] = new EntityReference("mcs_services", new Guid(tsapresent));

                newGroupResource["cvt_relatedsiteid"] = new EntityReference("mcs_site", McsHelper.getEntRefID("mcs_relatedsiteid"));
                newGroupResource["cvt_relatedresourceid"] = new EntityReference("mcs_resource", primaryEntityId);
                newGroupResource["cvt_name"] = McsHelper.getStringValue("mcs_name");

                Logger.WriteDebugMessage("About to create provider resource group record");
                OrganizationService.Create(newGroupResource);
                Logger.WriteDebugMessage("created provider resource group record");
            }
            catch (FaultException<OrganizationServiceFault> ex)
            {
                Logger.WriteToFile("Error creating new Group Resource:" + ex.Message);
                throw new InvalidPluginExecutionException(ex.Message);
            }
            catch (Exception ex)
            {
                Logger.WriteToFile("Error creating new Group Resource:" + ex.Message);
                throw new InvalidPluginExecutionException(ex.Message);
            }
        }
        internal void CreateSysResource(Guid thisId)
        {
            //Starting the creation of a System Resource which will be associated with target MCS Resource. 
            Logger.WriteDebugMessage("starting CreateSysResource");

            var whereami = "top";

            using (var srv = new Xrm(OrganizationService))
            {
                try
                {
                    Logger.setMethod = "Create System Resource";

                    var thisResource = srv.mcs_resourceSet.FirstOrDefault(i => i.Id == thisId);
                    //Creating a new System Resource ('Equipment') , and assigning attribute values. 
                    var newSystemResource = new Entity("equipment");
                    newSystemResource["name"] = McsHelper.getStringValue("mcs_name");

                    var ResourceTypeValue = thisResource.mcs_Type.Value;

                    if (ResourceTypeValue == 251920000)
                    {
                        if (thisResource.cvt_vistacapacity != null)
                        {
                            newSystemResource["cvt_capacity"] = thisResource.cvt_vistacapacity.Value;
                        }                       
                    }
                    else
                    {
                        newSystemResource["cvt_capacity"] = thisResource.cvt_capacity;
                    }
                    newSystemResource["cvt_type"] = thisResource.mcs_Type;

                    if (thisResource.mcs_BusinessUnitId != null)
                    {
                        newSystemResource["businessunitid"] = thisResource.mcs_BusinessUnitId;
                    }
                    else
                    {
                        var relatedSite = srv.mcs_siteSet.FirstOrDefault(i => i.Id == McsHelper.getEntRefID("mcs_relatedsiteid"));
                        newSystemResource["businessunitid"] = relatedSite.mcs_BusinessUnitId;
                    }

                  //  newSystemResource["businessunitid"] = new EntityReference("businessunit", McsHelper.getEntRefID("owningbusinessunit"));
                    Logger.WriteDebugMessage("about to get site");
                    whereami = "about to get site";
                    var mcsSite = srv.mcs_siteSet.FirstOrDefault(i => i.Id == McsHelper.getEntRefID("mcs_relatedsiteid"));

                    //Checking to make sure we have an MCS Site. 
                    whereami = "just after get site";
                    if (mcsSite == null)
                    {
                        whereami = "no site";
                        Logger.WriteToFile("Did not get the site");
                        return;
                    }
                                                         
                    whereami = "got site";
                    newSystemResource["siteid"] = new EntityReference("site", mcsSite.mcs_RelatedActualSiteId.Id);
                    newSystemResource["timezonecode"] = mcsSite.mcs_TimeZone;
                    newSystemResource["mcs_relatedresource"] = new EntityReference("mcs_resource", thisId);
                    whereami = "about to create";                  
                    
                    //The System Resource is created given all attributes were succesfully assigned. 
                    var newSysResource = OrganizationService.Create(newSystemResource);
                    whereami = "created";
                    Logger.WriteDebugMessage("System Resource Added:" + newSysResource.ToString());

                    //Here we are updating the target MCS Resource that was created, and assigning the System Resource we just created to the lookup so they are assoicated. 
                    Logger.setMethod = "Update Mcs Resource";
                    //var updateResource = new Entity("mcs_resource") {Id = thisId};
                    var updateResource = new mcs_resource()
                    {
                        Id = thisId
                    };
                    updateResource.mcs_relatedResourceId = new EntityReference(Equipment.EntityLogicalName, newSysResource);

                    //Update TSS Resource's VISN field
                    if ((thisResource.mcs_BusinessUnitId == null) || (thisResource.mcs_BusinessUnitId.Id != mcsSite.mcs_BusinessUnitId.Id))
                    {
                        updateResource.mcs_BusinessUnitId = mcsSite.mcs_BusinessUnitId;
                        Logger.WriteDebugMessage("Updated VISN field");
                    }
                    //Update TSS Resource's Facility field
                    var thisResourceFacility = (thisResource.mcs_Facility != null) ? thisResource.mcs_Facility.Id : Guid.Empty;
                    var siteFacility = (mcsSite.mcs_FacilityId != null) ? mcsSite.mcs_FacilityId.Id : Guid.Empty;


                    if ((thisResourceFacility == Guid.Empty && siteFacility != Guid.Empty) || (thisResourceFacility != siteFacility))
                    {
                        updateResource.mcs_Facility = mcsSite.mcs_FacilityId;
                        Logger.WriteDebugMessage("Updated Facility field");
                    } 
                 
                    OrganizationService.Update(updateResource);
                    whereami = "Updated";
                    Logger.WriteDebugMessage("MCS Resource Updated");

                    //Updating the default calendar hours for the System Resource that was created. 
                    Logger.WriteGranularTimingMessage("Starting ChangeNewlyCreatedCalendar");
                    CvtHelper.ChangeNewlyCreatedCalendar(newSysResource, OrganizationService, Logger, McsSettings);
                    Logger.WriteGranularTimingMessage("Ending ChangeNewlyCreatedCalendar");

                    Logger.setMethod = "Create System Resource";
                    Logger.WriteDebugMessage("Calendar Updated");
                    whereami = "Calendar Updated";

                    ColumnSet attributes = new ColumnSet(new string[] { "cvt_primaryteam" });

                    // Retrieve the system user's primary team
                    Entity initiatingUserEntity = OrganizationService.Retrieve("systemuser", PluginExecutionContext.InitiatingUserId, attributes);
                    if (initiatingUserEntity.Attributes.Contains("cvt_primaryteam"))
                    {
                        var myTeam = (EntityReference)initiatingUserEntity["cvt_primaryteam"];
                        AssignRequest assignRequest = new AssignRequest()
                        {
                            Assignee = new EntityReference
                            {
                                LogicalName = Team.EntityLogicalName,
                                Id = myTeam.Id
                            },

                            Target = new EntityReference("mcs_resource", thisId)
                        };

                        OrganizationService.Execute(assignRequest);

                    }
                    whereami = "Assigned";
                }
                catch (FaultException<OrganizationServiceFault> ex)
                {
                    Logger.WriteToFile(whereami + ex.Message);
                    throw new InvalidPluginExecutionException(ex.Message);
                }
                catch (Exception ex)
                {
                    Logger.WriteToFile(whereami + ex.Message);
                    throw new InvalidPluginExecutionException(ex.Message);
                }
            }
        }
        #endregion

        #region Additional Interface methods
        public override string McsSettingsDebugField
        {
            get { return "mcs_resourceplugin"; }
        }

        public override Entity GetPrimaryEntity()
        {
            return (Entity)PluginExecutionContext.InputParameters["Target"];
        }

        public override Entity GetSecondaryEntity()
        {
            return (Entity)PluginExecutionContext.PostEntityImages["post"];
        }
        #endregion
    }
}
