﻿using System;
using System.ServiceModel;
using Microsoft.Xrm.Sdk;
using mcsScheduling;
using System.Linq;

namespace MCSPlugins
{
    public class McsResourceDeletePreStageRunner : PluginRunner
    {
        #region Constructor
        public McsResourceDeletePreStageRunner(IServiceProvider serviceProvider)
            : base(serviceProvider)
        {
        }
        #endregion

        #region Internal Methods/Properties
        internal void Execute()
        {
            try
            {
                Logger.setMethod = "Execute";
                Logger.WriteTxnTimingMessage("Starting");
                if (PluginExecutionContext.InputParameters.Contains("Target") &&
                    PluginExecutionContext.InputParameters["Target"] is EntityReference)
                {
                    Logger.WriteDebugMessage(String.Format("At the top of:{0}", GetType()));                                                  
                    Logger.WriteGranularTimingMessage("Starting Delete Resource");
                    CheckComponents(PluginExecutionContext.PrimaryEntityId);
                    DeleteResource(PluginExecutionContext.PrimaryEntityId);
                    Logger.WriteGranularTimingMessage("Ending Delete Resource");
                }
                Logger.setMethod = "Execute";
                Logger.WriteTxnTimingMessage("Ending");
            }
            catch (FaultException<OrganizationServiceFault> ex)
            {
                Logger.WriteToFile(ex.Message);
                throw new InvalidPluginExecutionException(McsSettings.getUnexpectedErrorMessage + ":" + ex.Message);
            }
            catch (Exception ex)
            {
                if (ex.Message.StartsWith("custom"))
                {
                    Logger.WriteDebugMessage(ex.Message.Substring(6));
                    throw new InvalidPluginExecutionException(ex.Message.Substring(6));
                }
                else
                {
                    Logger.setMethod = "Execute";
                    Logger.WriteToFile(ex.Message);
                    throw new InvalidPluginExecutionException(McsSettings.getUnexpectedErrorMessage + ":" + ex.Message);
                }
            }
        }

        internal void CheckComponents(Guid thisId)
        {
            Logger.setMethod = "CheckComponents";
            Logger.WriteDebugMessage("Starting CheckComponents");
            using (var srv = new Xrm(OrganizationService))
            {             
                //Checking to see if any children Components Exists for the target MCS Resource being deleted. 
                var childComponents = srv.cvt_componentSet.FirstOrDefault(c => c.cvt_relatedresourceid.Id == thisId);

                //If a System Resource does not exist, we will return because the Plugin will have nothing to delete. 
                if (childComponents != null)
                {
                    Logger.WriteDebugMessage("Detected Components, stopping delete");
                    throw new InvalidPluginExecutionException("customDelete cancelled. Retry delete after you have re-assigned all related Components to another Technology Resource.");  
                }
            }
            Logger.WriteDebugMessage("Ending CheckComponents");
        }
        internal void DeleteResource(Guid thisId)
        {
            Logger.WriteDebugMessage("Starting Delete System Resource");
            using (var srv = new Xrm(OrganizationService))
            {
                //Starting Validation Checks before deleting the MCS Resource & System Resource. 
                Logger.setMethod = "Delete Sytem Resource";
                //Checking to see if a System Resource Exists for the target MCS Resource being deleted. 
                var systemEquip = srv.EquipmentSet.FirstOrDefault(i => i.mcs_relatedresource.Id == thisId);

                //If a System Resource does not exist, we will return because the Plugin will have nothing to delete. 
                if (systemEquip == null)
                    return;
                Logger.WriteDebugMessage("Got System Resource(facility/equip):" + systemEquip.Name);

                //Checking to see if any Patient Site Resources have this target MCS Resource associated with it.
                var patResource = srv.cvt_patientresourcegroupSet.FirstOrDefault(i => i.cvt_RelatedResourceId.Id == thisId);                                  
                //If an associated Patient Site Resource does exist, we will throw an exception with a message and stop the delete of the MCS Resource, to prevent orphan data. 
                if (patResource != null)
                {
                    Logger.WriteDebugMessage("Patient Resource Exists:" + patResource.cvt_name);                                  
                    throw new InvalidPluginExecutionException("customPlease check Left Nav on Resource form for Patient Site Resources that this Resource is associated with. Resource cannot be deleted until associations are removed.");                      
                }
                //Checking to see if any Provider Site Resources have this target MCS Resource associated with it. 
                var proResource = srv.cvt_providerresourcegroupSet.FirstOrDefault(i => i.cvt_RelatedResourceId.Id == thisId);
                //If an associated Provider Site Resource does exist, we will throw an exception with a message and stop the delete of the MCS Resource, to prevent orphan data.
                if (proResource != null)
                {
                    Logger.WriteDebugMessage("Provider Resource Exists:" + proResource.cvt_name);                                     
                    throw new InvalidPluginExecutionException("customPlease check Left Nav on Resource form for Provider Site Resources that this Resource is associated with. Resource cannot be deleted until associations are removed.");
                }
                //Checking to see if any Group Resources have this target MCS Resource associated with it.
                var groupResource = srv.mcs_groupresourceSet.FirstOrDefault(i => i.mcs_RelatedResourceId.Id == thisId);
                //If an associated Group Resource does exist, we will throw an exception with a message and stop the delete of the MCS Resource, to prevent orphan data.
                if (groupResource != null)
                {
                    Logger.WriteDebugMessage("Group Resource Exists:" + groupResource.mcs_name);                                    
                    throw new InvalidPluginExecutionException("customPlease check Left Nav on Resource form for Group Resources that this Resource is associated with. Resource cannot be deleted until associations are removed.");
                }
                var parties = srv.ActivityPartySet.Where(ap => ap.PartyId.Id == systemEquip.Id).ToList();
                if (parties.Count > 0)
                    throw new InvalidPluginExecutionException(String.Format("TSS Resource Delete not performed, Please Deactivate instead.  Could not Delete related Equipment ({0}) because it exists on {1} activity record(s).", systemEquip.Name, parties.Count().ToString()));
                else
                {
                    OrganizationService.Delete(systemEquip.LogicalName, systemEquip.Id);
                    Logger.WriteDebugMessage("System Resource Deleted");
                }
            }
        }

        public override string McsSettingsDebugField
        {
            get { return "mcs_resourceplugin"; }
        }

        public override Entity GetPrimaryEntity()
        {
            return (Entity)PluginExecutionContext.PreEntityImages["pre"];
        }

        public override Entity GetSecondaryEntity()
        {
            return (Entity)PluginExecutionContext.PreEntityImages["pre"];
        }
    
        #endregion
    }
}
