﻿using System;
using System.ServiceModel;
using Microsoft.Xrm.Sdk;
using mcsScheduling;
using System.Linq;
using Microsoft.Xrm.Sdk.Query;
using MCSShared;
using System.Text;


namespace MCSPlugins
{
    public class McsResourceUpdatePostStageRunner : PluginRunner
    {
        #region Constructor
        public McsResourceUpdatePostStageRunner(IServiceProvider serviceProvider)
            : base(serviceProvider)
        {
        }       
        #endregion
        
        #region Internal Methods
        /// <summary>
        /// This plugin runs after the mcs_Resource has been updated.
        ///It accomplishes the following:
        ///1.Checks to see if the mcs_Resource has the Resource Group GUID text field, and will run the UpdateResourceGroup method (In MCS Resource Group Plugin Update PS Runner), if it has a GUID. If the Resource Group GUID text field is null, the method does not run. 
        /// 2.Updates all attributes for the System Resource that is associated with the mcs_Resource. 
        /// </summary>
        internal void Execute()
        {
            try
            {
                if (PluginExecutionContext.Depth > 2) { return; }

                Logger.setMethod = "Execute";
                Logger.WriteTxnTimingMessage("Starting");

                if (PluginExecutionContext.InputParameters.Contains("Target") && PluginExecutionContext.InputParameters["Target"] is Entity)
                {
                    Logger.WriteDebugMessage(String.Format("At the top of:{0}", GetType()));

                    var entity = (Entity)PluginExecutionContext.InputParameters["Target"];
                    var thisResource = entity.ToEntity<mcs_resource>();

                    //We will check these Boolean fields first to see if they have been flagged for Update / Replace / Delete. 
                    bool updateResourceConnections = false;
                    bool replaceResourceConnections = false;
                    bool deleteResourceConnections = false;
                    if (thisResource.cvt_updateresourceconnections != null) { updateResourceConnections = thisResource.cvt_updateresourceconnections.Value; }
                    if (thisResource.cvt_replaceresourceconnections != null) { replaceResourceConnections = thisResource.cvt_replaceresourceconnections.Value; }
                    if (thisResource.cvt_deleteresourceconnections != null) { deleteResourceConnections = thisResource.cvt_deleteresourceconnections.Value; }

                    var ReplacementResource = thisResource.cvt_replacementresource;
                    //Update self with new Facility and VISN.
                    CvtHelper.AlignLocations(thisResource, OrganizationService, Logger);

                    if (updateResourceConnections == true)
                    {
                        Logger.WriteGranularTimingMessage("Starting UpdateResourceConnections");
                        UpdateResourceConnections(PluginExecutionContext.PrimaryEntityId);
                        Logger.WriteGranularTimingMessage("Ending UpdateResourceConnections");
                    }
                    if (ReplacementResource != null && replaceResourceConnections == true)
                    {
                        var ReplacementResourceId = McsHelper.getEntRefID("cvt_replacementresource");
                        Logger.WriteGranularTimingMessage("Starting ReplaceResourceConnections");
                        ReplaceResourceConnections(PluginExecutionContext.PrimaryEntityId, ReplacementResourceId);
                        Logger.WriteGranularTimingMessage("Ending ReplaceResourceConnections");
                    }
                    if (deleteResourceConnections == true)
                    {
                        Logger.WriteGranularTimingMessage("Starting DeleteResourceConnections");
                        DeleteResourceConnections(PluginExecutionContext.PrimaryEntityId);
                        Logger.WriteGranularTimingMessage("Ending DeleteResourceConnections");
                    }

                    Logger.WriteGranularTimingMessage("Starting UpdateResourceGroup");
                    UpdateResourceGroup(PluginExecutionContext.PrimaryEntityId);
                    Logger.WriteGranularTimingMessage("Ending UpdateResourceGroup");

                    if (thisResource.mcs_relatedResourceId != null)
                        UpdateSystemResource(PluginExecutionContext.PrimaryEntityId, thisResource);                    
                    else
                    {
                        var resource = (mcs_resource)OrganizationService.Retrieve(mcs_resource.EntityLogicalName, PluginExecutionContext.PrimaryEntityId, new ColumnSet(true));
                        if (resource.mcs_relatedResourceId != null)
                            UpdateSystemResource(resource.Id, resource);
                        else
                            Logger.WriteDebugMessage("Not Doing Update of Sys Resource");
                    }
                    if (entity.Attributes.Contains("mcs_relatedsiteid") && PluginExecutionContext.Depth == 1){
                        var error = ValidateMoveResource(entity.Id);
                        if (error != String.Empty)
                            throw new InvalidPluginExecutionException(error);
                    }
                }

                Logger.setMethod = "Execute";
                Logger.WriteTxnTimingMessage("Ending");
            }
            catch (FaultException<OrganizationServiceFault> ex)
            {
                Logger.WriteToFile(ex.Message);
                throw new InvalidPluginExecutionException(ex.Message);
            }
            catch (Exception ex)
            {
                if (ex.Message.StartsWith("custom"))
                {
                    Logger.WriteDebugMessage(ex.Message.Substring(6));
                    throw new InvalidPluginExecutionException(ex.Message.Substring(6));
                }
                else
                {
                    Logger.setMethod = "Execute";
                    Logger.WriteToFile(ex.Message);
                    throw new InvalidPluginExecutionException(ex.Message);
                }
            }
        }

        /// <summary>
        /// Check the TSS Resource for Related records and throw an exception if any exist - look for active TSAs, Appts, SAs, and Resource Groups
        /// </summary>
        internal string ValidateMoveResource(Guid resourceId)
        {
            //instantiate return variable
            var errorMessage = String.Empty;
            var equipmentId = McsHelper.getEntRefID("mcs_relatedresourceid");
            using (var srv = new Xrm(OrganizationService))
            {
                //Get All Relevant Relationships for this resource (patGroup is active, tsa is active; same for ProGroup; Resource Group is active, activity scheduled is future and is in scheduled status)
                var relevantPats = (from patGroup in srv.cvt_patientresourcegroupSet
                                    join tsa in srv.mcs_servicesSet on patGroup.cvt_RelatedTSAid.Id equals tsa.mcs_servicesId
                                    where patGroup.cvt_RelatedResourceId.Id == resourceId
                                    where patGroup.statecode.Value == cvt_patientresourcegroupState.Active
                                    where tsa.statecode.Value == mcs_servicesState.Active
                                    select new
                                    {
                                        patGroup.Id //Only care about counts, so just select id column
                                    }).ToList();

                var relevantPros = (from provGroup in srv.cvt_providerresourcegroupSet
                                    join tsa in srv.mcs_servicesSet on provGroup.cvt_RelatedTSAid.Id equals tsa.mcs_servicesId
                                    where provGroup.cvt_RelatedResourceId.Id == resourceId
                                    where provGroup.statecode.Value == cvt_providerresourcegroupState.Active
                                    where tsa.statecode.Value == mcs_servicesState.Active
                                    select new
                                    {
                                        provGroup.Id //Only care about counts, so just select id column
                                    }).ToList();

                var relevantGroups = (from groupRes in srv.mcs_groupresourceSet
                                      join resGroup in srv.mcs_resourcegroupSet on groupRes.mcs_relatedResourceGroupId.Id equals resGroup.Id
                                      where groupRes.mcs_RelatedResourceId.Id == resourceId
                                      where resGroup.statecode.Value == mcs_resourcegroupState.Active
                                      where groupRes.statecode.Value == mcs_groupresourceState.Active
                                      select new
                                      {
                                          id = groupRes.mcs_relatedResourceGroupId.Id //need resource group ID for later query
                                      }).ToList();

                var tsas = relevantPros.Count() + relevantPats.Count();
                if (relevantGroups != null && relevantGroups.Count() > 0)
                {
                    var proGroups = srv.cvt_providerresourcegroupSet.Where(p => p.cvt_RelatedTSAid != null && p.cvt_RelatedResourceGroupid != null);
                    var groupProTSAs = (from groups in relevantGroups
                                        join provGroups in proGroups on groups.id equals provGroups.cvt_RelatedResourceGroupid.Id
                                        join tsa in srv.mcs_servicesSet on provGroups.cvt_RelatedTSAid.Id equals tsa.Id
                                        where provGroups.statecode.Value == cvt_providerresourcegroupState.Active
                                        where tsa.statecode.Value == mcs_servicesState.Active
                                        select new
                                        {
                                            tsa.Id
                                        }).ToList();

                    var patGroupSet = srv.cvt_patientresourcegroupSet.Where(p => p.cvt_RelatedTSAid != null && p.cvt_RelatedResourceGroupid != null);
                    var groupPatTSAs = (from groups in relevantGroups
                                        join patGroups in patGroupSet on groups.id equals patGroups.cvt_RelatedResourceGroupid.Id
                                        join tsa in srv.mcs_servicesSet on patGroups.cvt_RelatedTSAid.Id equals tsa.Id
                                        where patGroups.statecode.Value == cvt_patientresourcegroupState.Active
                                        where tsa.statecode.Value == mcs_servicesState.Active
                                        select new
                                        {
                                            tsa.Id
                                        }).ToList();
                    tsas += groupProTSAs.Count() + groupPatTSAs.Count();
                }

                var activities = (from ap in srv.ActivityPartySet
                                  join activity in srv.ActivityPointerSet on ap.ActivityId.Id equals activity.Id
                                  where ap.PartyId.Id == equipmentId
                                  where (activity.StateCode == ActivityPointerState.Scheduled || activity.StateCode == ActivityPointerState.Open)
                                  where (activity.ScheduledStart.Value > DateTime.Now.ToUniversalTime()) //activity is in future
                                  select new
                                  {
                                      activity.Id //Only care about counts, so just select id column
                                  }).ToList();

                if (activities.Count() + relevantGroups.Count() + tsas > 0)
                {
                    errorMessage = String.Format("You have {0} Activities, {1} TSAs, and {2} Resource Groups related to this TSS Resource.  ", activities.Count(), tsas, relevantGroups.Count());
                    if (activities.Count() > 0)
                        errorMessage += "\nYou cannot change the site for this resource until you have cancelled or rescheduled all of its appointments and service activities";
                    if (tsas > 0)
                        errorMessage += "\nYou cannot change the site for this resource until you have removed or replaced it on all TSAs (with the \"Replace Resource\" or \"Remove All Connections\" button)";
                    if (relevantGroups.Count() > 0)
                        errorMessage += "\nyou cannot change the site for this resource until you have removed it from all existing resource groups";
                }
            }
            return errorMessage;
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="thisId"></param>
        internal void UpdateResourceGroup(Guid thisId)
        {          
            //Checks to see if the Resource Group GUID Text Field has a value, and will run the updateResourceGroup method in Resoure Group Update Plugin if not null. 
            var resourceGroupGuid = McsHelper.getStringValue("mcs_resourcegroupguid");
            var updateResourceGroup = resourceGroupGuid != null;

            if (!updateResourceGroup)
                return;
            Logger.WriteDebugMessage("Updating Resource Group");
        }

        internal void UpdateSystemResource(Guid thisId, mcs_resource resource)
        {
            Logger.setMethod = "UpdateSystemResource";
            Logger.WriteGranularTimingMessage("Starting UpdateSystemResource");
            using (var srv = new Xrm(OrganizationService))
            {
                try
                {                  
                    //Grabs all the attributes needed to update the associated System Resource of the mcs_Resource that was just updated. 
                    var systemEquip = srv.EquipmentSet.FirstOrDefault(i => i.Id == resource.mcs_relatedResourceId.Id);
                    if (systemEquip == null) { return; }

                    var thisResource = srv.mcs_resourceSet.FirstOrDefault(r => r.Id == thisId);
                    var updateSysResource = new Entity("equipment") {Id = systemEquip.Id};
                    updateSysResource["name"] = McsHelper.getStringValue("mcs_name");
                    updateSysResource["cvt_type"] = thisResource.mcs_Type;
                    updateSysResource["mcs_relatedresource"] = new EntityReference("mcs_resource", thisId);
                    updateSysResource["businessunitid"] = new EntityReference("businessunit", McsHelper.getEntRefID("owningbusinessunit"));

                    var ResourceTypeValue = thisResource.mcs_Type.Value;
                    if (ResourceTypeValue == 251920000)
                    {
                        if (thisResource.cvt_vistacapacity != null)
                        {
                            updateSysResource["cvt_capacity"] = thisResource.cvt_vistacapacity.Value;
                            ChangeNewlyCreatedCalendarUpdate(systemEquip.Id);
                            //CvtHelper.ChangeNewlyCreatedCalendar(systemEquip.Id, OrganizationService, Logger, McsSettings);
                        }                       
                    }
                    else
                        updateSysResource["cvt_capacity"] = thisResource.cvt_capacity;
                    
                    var mcsSite = srv.mcs_siteSet.FirstOrDefault(i => i.Id == McsHelper.getEntRefID("mcs_relatedsiteid"));
                    if (mcsSite == null) 
                        return;

                    Logger.WriteDebugMessage("got mcssite: " + mcsSite.mcs_RelatedActualSiteId.Name);
                    updateSysResource["siteid"] = new EntityReference("site", mcsSite.mcs_RelatedActualSiteId.Id);
                    updateSysResource["timezonecode"] = mcsSite.mcs_TimeZone;

                    //Updates System Resource
                    OrganizationService.Update(updateSysResource);
                    Logger.WriteDebugMessage("System Resource Updated");
                    Logger.WriteGranularTimingMessage("Ending UpdateSytemResource");
                }
                catch (FaultException<OrganizationServiceFault> ex)
                {
                    Logger.WriteToFile(ex.Message);
                    throw new InvalidPluginExecutionException(ex.Message);
                }
                catch (Exception ex)
                {
                    Logger.WriteToFile(ex.Message);
                    throw new InvalidPluginExecutionException(ex.Message);
                }                
            }
        }

        internal void UpdateResourceConnections(Guid thisId)
        {
            Logger.setMethod = "UpdateResourceConnections";
            using (var srv = new Xrm(OrganizationService))
            {
                try
                {                    
                    Logger.WriteDebugMessage("Looking up Resource. ");
                    mcs_resource thisResource = (mcs_resource)OrganizationService.Retrieve("mcs_resource", thisId, new ColumnSet(true));

                    if (thisResource == null) { return; }
                    Logger.WriteDebugMessage("Got Resource");

                    //Going through Pat Site Resources that need to be updated. 
                    var getPatResources = from patGroup in srv.cvt_patientresourcegroupSet
                                          where patGroup.cvt_RelatedResourceId.Id == thisId
                                          where patGroup.statecode == 0
                                          select new
                                          {
                                              patGroup.Id,
                                              patGroup.cvt_RelatedTSAid
                                          };
                    foreach (var patGroup in getPatResources)
                    {
                        var updatePatGroup = new Entity("cvt_patientresourcegroup") { Id = patGroup.Id };
                        //updatePatGroup["cvt_relatedresourceid"] = new EntityReference("mcs_resource", thisId);
                        updatePatGroup["cvt_name"] = thisResource.mcs_name;
                        updatePatGroup["cvt_type"] = thisResource.mcs_Type;
                        updatePatGroup["cvt_relatedsiteid"] = new EntityReference("mcs_site", thisResource.mcs_RelatedSiteId.Id);
                        updatePatGroup["cvt_capacityrequired"] = thisResource.cvt_capacity;
                        updatePatGroup["cvt_roomcapacity"] = thisResource.cvt_capacity;
                        updatePatGroup["cvt_resourcespecguid"] = thisResource.mcs_resourcespecguid;
                        updatePatGroup["cvt_constraintgroupguid"] = thisResource.mcs_constraintgroupguid;
                        OrganizationService.Update(updatePatGroup);
                        Logger.WriteDebugMessage("Pat Group Updated");

                        if (patGroup.cvt_RelatedTSAid != null)
                            updateTSAfields(patGroup.cvt_RelatedTSAid.Id);
                    }

                    //Going through Prov Site Resources that need to be updated
                    var getProvResources = from provGroup in srv.cvt_providerresourcegroupSet
                                           where provGroup.cvt_RelatedResourceId.Id == thisId
                                           where provGroup.statecode == 0
                                           select new
                                           {
                                               provGroup.Id,
                                               provGroup.cvt_RelatedTSAid,
                                               provGroup.cvt_RelatedMasterTSAId
                                           };
                    foreach (var provGroup in getProvResources)
                    {
                        var updateProvGroup = new Entity("cvt_providerresourcegroup") { Id = provGroup.Id };
                        //updateProvGroup["cvt_relatedresourceid"] = new EntityReference("mcs_resource", thisId);
                        updateProvGroup["cvt_name"] = thisResource.mcs_name;
                        updateProvGroup["cvt_type"] = thisResource.mcs_Type;
                        updateProvGroup["cvt_relatedsiteid"] = new EntityReference("mcs_site", thisResource.mcs_RelatedSiteId.Id);
                        updateProvGroup["cvt_capacityrequired"] = thisResource.cvt_capacity;
                        updateProvGroup["cvt_roomcapacity"] = thisResource.cvt_capacity;
                        updateProvGroup["cvt_resourcespecguid"] = thisResource.mcs_resourcespecguid;
                        updateProvGroup["cvt_constraintgroupguid"] = thisResource.mcs_constraintgroupguid;
                        OrganizationService.Update(updateProvGroup);
                        Logger.WriteDebugMessage("Prov Group Updated");

                        if (provGroup.cvt_RelatedTSAid != null)
                            updateTSAfields(provGroup.cvt_RelatedTSAid.Id);
                        if (provGroup.cvt_RelatedMasterTSAId != null)
                            updateMTSAfields(provGroup.cvt_RelatedMasterTSAId.Id);
                    }

                    //Going through Group Resources that need to be updated
                    var getGroupResources = from groupResource in srv.mcs_groupresourceSet
                                            where groupResource.mcs_RelatedResourceId.Id == thisId
                                            where groupResource.statecode == 0
                                            select new
                                            {
                                                groupResource.Id,
                                                groupResource.mcs_relatedResourceGroupId
                                            };
                    foreach (var groupResource in getGroupResources)
                    {
                        var updateGroupResource = new Entity("mcs_groupresource") { Id = groupResource.Id };
                        var relatedResourceGroup = srv.mcs_resourcegroupSet.FirstOrDefault(i => i.Id == groupResource.mcs_relatedResourceGroupId.Id);
                       // updateGroupResource["mcs_relatedresourceid"] = new EntityReference("mcs_resource", thisId);
                        updateGroupResource["mcs_name"] = thisResource.mcs_name;
                        updateGroupResource["mcs_type"] = thisResource.mcs_Type;
                        updateGroupResource["mcs_relatedsiteid"] = new EntityReference("mcs_site", thisResource.mcs_RelatedSiteId.Id);
                        updateGroupResource["cvt_capacity"] = thisResource.cvt_capacity;
                        OrganizationService.Update(updateGroupResource);
                        Logger.WriteDebugMessage("Group Resource Updated");

                        //Update the Resource Group here. 
                        UpdateRG(relatedResourceGroup.Id);
                        Logger.WriteDebugMessage("Resource Group Updated");

                        //now we need to look for Pat / Pro Site Resources related to this Resource Group. And update the TSA's related to them. 
                        //Going through Pat Site Resources that need to be deleted. 
                        var getRGPatResources = from RGpatGroup in srv.cvt_patientresourcegroupSet
                                                where RGpatGroup.cvt_RelatedResourceGroupid.Id == relatedResourceGroup.Id
                                                where RGpatGroup.statecode == 0
                                                select new
                                                {
                                                    RGpatGroup.Id,
                                                    RGpatGroup.cvt_RelatedTSAid
                                                };
                        foreach (var RGpatGroup in getRGPatResources)
                        {
                            if (RGpatGroup.cvt_RelatedTSAid != null)
                                updateTSAfields(RGpatGroup.cvt_RelatedTSAid.Id);
                        }

                        var getRGProvResources = from RGprovGroup in srv.cvt_providerresourcegroupSet
                                                 where RGprovGroup.cvt_RelatedResourceGroupid.Id == relatedResourceGroup.Id
                                                 where RGprovGroup.statecode == 0
                                                 select new
                                                 {
                                                     RGprovGroup.Id,
                                                     RGprovGroup.cvt_RelatedTSAid,
                                                     RGprovGroup.cvt_RelatedMasterTSAId
                                                 };
                        foreach (var RGprovGroup in getRGProvResources)
                        {
                            if (RGprovGroup.cvt_RelatedTSAid != null)
                                updateTSAfields(RGprovGroup.cvt_RelatedTSAid.Id);
                            if (RGprovGroup.cvt_RelatedMasterTSAId != null)
                                updateMTSAfields(RGprovGroup.cvt_RelatedMasterTSAId.Id);
                        }
                    }

                    var updateResource = new Entity("mcs_resource") { Id = thisId };
                    updateResource["cvt_updateresourceconnections"] = false;
                    OrganizationService.Update(updateResource);
                    Logger.WriteDebugMessage("cvt_updateresourceconnections field updated");
                }
                catch (FaultException<OrganizationServiceFault> ex)
                {
                    Logger.WriteToFile(ex.Message);
                    throw new InvalidPluginExecutionException(ex.Message);
                }
                catch (Exception ex)
                {
                    Logger.WriteToFile(ex.Message);
                    throw new InvalidPluginExecutionException(ex.Message);
                }
            }
        }

        internal void ReplaceResourceConnections(Guid thisId, Guid ReplacementResourceId)
        {
            Logger.setMethod = "ReplaceResourceConnections";
            using (var srv = new Xrm(OrganizationService))
            {
                try
                {
                    var replacementResource = srv.mcs_resourceSet.FirstOrDefault(i => i.Id == ReplacementResourceId);
                    if (replacementResource == null) { return; }
                    Logger.WriteDebugMessage("Got Replacement Resource");

                    //Going through Pat Site Resources that need to be updated. 
                    var getPatResources = from patGroup in srv.cvt_patientresourcegroupSet
                                          where patGroup.cvt_RelatedResourceId.Id == thisId
                                          where patGroup.statecode == 0
                                          select new
                                          {
                                              patGroup.Id,
                                              patGroup.cvt_RelatedTSAid
                                          };
                    foreach (var patGroup in getPatResources)
                    {
                        var updatePatGroup = new Entity("cvt_patientresourcegroup") { Id = patGroup.Id };
                        updatePatGroup["cvt_relatedresourceid"] = new EntityReference("mcs_resource", ReplacementResourceId);
                        updatePatGroup["cvt_name"] = replacementResource.mcs_name;
                        updatePatGroup["cvt_type"] = replacementResource.mcs_Type;
                        updatePatGroup["cvt_relatedsiteid"] = new EntityReference("mcs_site", replacementResource.mcs_RelatedSiteId.Id);
                        updatePatGroup["cvt_capacityrequired"] = replacementResource.cvt_capacity;
                        updatePatGroup["cvt_roomcapacity"] = replacementResource.cvt_capacity;
                        updatePatGroup["cvt_resourcespecguid"] = replacementResource.mcs_resourcespecguid;
                        updatePatGroup["cvt_constraintgroupguid"] = replacementResource.mcs_constraintgroupguid;
                        OrganizationService.Update(updatePatGroup);
                        Logger.WriteDebugMessage("Pat Group Updated");

                        if (patGroup.cvt_RelatedTSAid != null)
                            updateTSAfields(patGroup.cvt_RelatedTSAid.Id);                     
                    }

                    //Going through Prov Site Resources that need to be updated
                    var getProvResources = from provGroup in srv.cvt_providerresourcegroupSet
                                          where provGroup.cvt_RelatedResourceId.Id == thisId
                                          where provGroup.statecode == 0
                                          select new
                                          {
                                              provGroup.Id,
                                              provGroup.cvt_RelatedTSAid,
                                              provGroup.cvt_RelatedMasterTSAId

                                          };
                    foreach (var provGroup in getProvResources)
                    {
                        var updateProvGroup = new Entity("cvt_providerresourcegroup") { Id = provGroup.Id };
                        updateProvGroup["cvt_relatedresourceid"] = new EntityReference("mcs_resource", ReplacementResourceId);
                        updateProvGroup["cvt_name"] = replacementResource.mcs_name;
                        updateProvGroup["cvt_type"] = replacementResource.mcs_Type;
                        updateProvGroup["cvt_relatedsiteid"] = new EntityReference("mcs_site", replacementResource.mcs_RelatedSiteId.Id);
                        updateProvGroup["cvt_capacityrequired"] = replacementResource.cvt_capacity;
                        updateProvGroup["cvt_roomcapacity"] = replacementResource.cvt_capacity;
                        updateProvGroup["cvt_resourcespecguid"] = replacementResource.mcs_resourcespecguid;
                        updateProvGroup["cvt_constraintgroupguid"] = replacementResource.mcs_constraintgroupguid;
                        OrganizationService.Update(updateProvGroup);
                        Logger.WriteDebugMessage("Prov Group Updated");

                        if (provGroup.cvt_RelatedTSAid != null)
                            updateTSAfields(provGroup.cvt_RelatedTSAid.Id);
                        if (provGroup.cvt_RelatedMasterTSAId != null)
                            updateMTSAfields(provGroup.cvt_RelatedMasterTSAId.Id);
                    }

                    //Going through Group Resources that need to be updated
                    var getGroupResources = from groupResource in srv.mcs_groupresourceSet
                                            where groupResource.mcs_RelatedResourceId.Id == thisId
                                            where groupResource.statecode == 0
                                           select new
                                           {
                                               groupResource.Id,
                                               groupResource.mcs_relatedResourceGroupId
                                           };
                    foreach (var groupResource in getGroupResources)
                    {                                                                       
                        var updateGroupResource = new Entity("mcs_groupresource") { Id = groupResource.Id };
                        var relatedResourceGroup = srv.mcs_resourcegroupSet.FirstOrDefault(i => i.Id == groupResource.mcs_relatedResourceGroupId.Id);
                        updateGroupResource["mcs_relatedresourceid"] = new EntityReference("mcs_resource", ReplacementResourceId);
                        updateGroupResource["mcs_name"] = replacementResource.mcs_name;
                        updateGroupResource["mcs_type"] = replacementResource.mcs_Type;
                        updateGroupResource["mcs_relatedsiteid"] = new EntityReference("mcs_site", replacementResource.mcs_RelatedSiteId.Id);
                        updateGroupResource["cvt_capacity"] = replacementResource.cvt_capacity;                    
                        OrganizationService.Update(updateGroupResource);
                        Logger.WriteDebugMessage("Group Resource Updated");

                        //Update the Resource Group here. 
                        UpdateRG(relatedResourceGroup.Id);
                        Logger.WriteDebugMessage("Resource Group Updated");

                        //now we need to look for Pat / Pro Site Resources related to this Resource Group. And update the TSA's related to them. 
                        //Going through Pat Site Resources that need to be deleted. 
                        var getRGPatResources = from RGpatGroup in srv.cvt_patientresourcegroupSet
                                                where RGpatGroup.cvt_RelatedResourceGroupid.Id == relatedResourceGroup.Id
                                                where RGpatGroup.statecode == 0
                                                select new
                                                {
                                                    RGpatGroup.Id,
                                                    RGpatGroup.cvt_RelatedTSAid
                                                };
                        foreach (var RGpatGroup in getRGPatResources)
                        {
                            if (RGpatGroup.cvt_RelatedTSAid != null)
                                updateTSAfields(RGpatGroup.cvt_RelatedTSAid.Id);
                        }

                        var getRGProvResources = from RGprovGroup in srv.cvt_providerresourcegroupSet
                                                 where RGprovGroup.cvt_RelatedResourceGroupid.Id == relatedResourceGroup.Id
                                                 where RGprovGroup.statecode == 0
                                                 select new
                                                 {
                                                     RGprovGroup.Id,
                                                     RGprovGroup.cvt_RelatedTSAid,
                                                     RGprovGroup.cvt_RelatedMasterTSAId
                                                 };
                        foreach (var RGprovGroup in getRGProvResources)
                        {
                            if (RGprovGroup.cvt_RelatedTSAid != null)
                                updateTSAfields(RGprovGroup.cvt_RelatedTSAid.Id);
                            if (RGprovGroup.cvt_RelatedMasterTSAId != null)
                            updateMTSAfields(RGprovGroup.cvt_RelatedMasterTSAId.Id);
                        }
                    }

                    var updateResource = new Entity("mcs_resource") { Id = thisId };
                    updateResource["cvt_replaceresourceconnections"] = false;
                    OrganizationService.Update(updateResource);
                    Logger.WriteDebugMessage("cvt_replaceresourceconnections field updated");
                }
                catch (FaultException<OrganizationServiceFault> ex)
                {
                    Logger.WriteToFile(ex.Message);
                    throw new InvalidPluginExecutionException(ex.Message);
                }
                catch (Exception ex)
                {
                    Logger.WriteToFile(ex.Message);
                    throw new InvalidPluginExecutionException(ex.Message);
                }
            }
        }

        internal void DeleteResourceConnections(Guid thisId)
        {
            Logger.setMethod = "DeleteResourceConnections";
            using (var srv = new Xrm(OrganizationService))
            {
                try
                {
                    //var thisResource = srv.mcs_resourceSet.FirstOrDefault(i => i.Id == thisId);

                    //Going through Pat Site Resources that need to be deleted. 
                    var getPatResources = from patGroup in srv.cvt_patientresourcegroupSet
                                          where patGroup.cvt_RelatedResourceId.Id == thisId
                                          where patGroup.statecode == 0
                                          select new
                                          {
                                              patGroup.Id,
                                              patGroup.cvt_RelatedTSAid
                                          };
                    foreach (var patGroup in getPatResources)
                    {
                        var deletePatGroup = new Entity("cvt_patientresourcegroup") { Id = patGroup.Id };
                        OrganizationService.Delete(deletePatGroup.LogicalName, patGroup.Id);
                        Logger.WriteDebugMessage("Pat Group Deleted");

                        if (patGroup.cvt_RelatedTSAid != null)
                            updateTSAfields(patGroup.cvt_RelatedTSAid.Id); 
             
                    }

                    //Going through Prov Site Resources that need to be deleted
                    var getProvResources = from provGroup in srv.cvt_providerresourcegroupSet
                                           where provGroup.cvt_RelatedResourceId.Id == thisId
                                           where provGroup.statecode == 0
                                           select new
                                           {
                                               provGroup.Id,
                                               provGroup.cvt_RelatedTSAid,
                                               provGroup.cvt_RelatedMasterTSAId
                                           };
                    foreach (var provGroup in getProvResources)
                    {
                        var deleteProvGroup = new Entity("cvt_providerresourcegroup") { Id = provGroup.Id };
                        OrganizationService.Delete(deleteProvGroup.LogicalName, provGroup.Id);
                        Logger.WriteDebugMessage("Prov Group Deleted");

                        if (provGroup.cvt_RelatedTSAid != null)
                            updateTSAfields(provGroup.cvt_RelatedTSAid.Id);
                        if (provGroup.cvt_RelatedMasterTSAId != null)
                            updateMTSAfields(provGroup.cvt_RelatedMasterTSAId.Id);
                    }

                    //Going through Group Resources that need to be deleted
                    var getGroupResources = from groupResource in srv.mcs_groupresourceSet
                                            where groupResource.mcs_RelatedResourceId.Id == thisId
                                            where groupResource.statecode == 0
                                            select new
                                            {
                                                groupResource.Id,
                                                groupResource.mcs_relatedResourceGroupId
                                            };
                    foreach (var groupResource in getGroupResources)
                    {
                        var deleteGroupResource = new Entity("mcs_groupresource") { Id = groupResource.Id };
                        var relatedResourceGroup = srv.mcs_resourcegroupSet.FirstOrDefault(i => i.Id == groupResource.mcs_relatedResourceGroupId.Id);
                        OrganizationService.Delete(deleteGroupResource.LogicalName, groupResource.Id);
                        Logger.WriteDebugMessage("Group Resource Deleted");

                        //After Deleting the Group Resource, we need to update the Resource Group.
                        UpdateRG(relatedResourceGroup.Id);
                        Logger.WriteDebugMessage("Resource Group Updated");

                        //now we need to look for Pat / Pro Site Resources related to this Resource Group. And update the TSA's related to them. 
                        //Going through Pat Site Resources that need to be deleted. 
                        var getRGPatResources = from RGpatGroup in srv.cvt_patientresourcegroupSet
                                              where RGpatGroup.cvt_RelatedResourceGroupid.Id == relatedResourceGroup.Id
                                              where RGpatGroup.statecode == 0
                                              select new
                                              {
                                                  RGpatGroup.Id,
                                                  RGpatGroup.cvt_RelatedTSAid
                                              };
                        foreach (var RGpatGroup in getRGPatResources)
                        {                        
                            if (RGpatGroup.cvt_RelatedTSAid != null)
                                updateTSAfields(RGpatGroup.cvt_RelatedTSAid.Id);                           
                        }

                        var getRGProvResources = from RGprovGroup in srv.cvt_providerresourcegroupSet
                                                where RGprovGroup.cvt_RelatedResourceGroupid.Id == relatedResourceGroup.Id
                                                where RGprovGroup.statecode == 0
                                                select new
                                                {
                                                    RGprovGroup.Id,
                                                    RGprovGroup.cvt_RelatedTSAid,
                                                    RGprovGroup.cvt_RelatedMasterTSAId
                                                };
                        foreach (var RGprovGroup in getRGProvResources)
                        {                           
                            if (RGprovGroup.cvt_RelatedTSAid != null)
                                updateTSAfields(RGprovGroup.cvt_RelatedTSAid.Id);
                            if (RGprovGroup.cvt_RelatedMasterTSAId != null)
                                updateMTSAfields(RGprovGroup.cvt_RelatedMasterTSAId.Id);                    
                        }
                    }
                    var updateResource = new Entity("mcs_resource") { Id = thisId };
                    updateResource["cvt_deleteresourceconnections"] = false;
                    OrganizationService.Update(updateResource);
                    Logger.WriteDebugMessage("cvt_deleteresourceconnections field updated");
                }
                catch (FaultException<OrganizationServiceFault> ex)
                {
                    Logger.WriteToFile(ex.Message);
                    throw new InvalidPluginExecutionException(ex.Message);
                }
                catch (Exception ex)
                {
                    Logger.WriteToFile(ex.Message);
                    throw new InvalidPluginExecutionException(ex.Message);
                }
            }
        }

        internal void UpdateRG(Guid thisId)
        {
            Logger.setMethod = "UpdateRG";
            Logger.WriteDebugMessage("starting UpdateResourceGroup");

            using (var srv = new Xrm(OrganizationService))
            {
                try
                {
                    var RG = srv.mcs_resourcegroupSet.FirstOrDefault(i => i.Id == thisId);
                    var resourceNames = String.Empty;
                    var count = 0;
                    var resources = CvtHelper.GetResources(thisId, srv, out resourceNames, out count);

                    var builder = CvtHelper.BuildConstraintsXML(resources);
                    var group = new ConstraintBasedGroup
                    {
                        Id = RG.mcs_RelatedResourceGroupId.Id,
                        Constraints = builder.ToString(),
                        Name = RG.mcs_name
                    };
                    Logger.WriteDebugMessage(String.Format("About to Update Constraint Group with {0} resources", count));
                    OrganizationService.Update(group);
                    var thisGroup = new mcs_resourcegroup()
                    {
                        Id = thisId,
                        cvt_resources = resourceNames
                    };
                    //updating resources string field for the resource group with new correct string
                    Logger.WriteDebugMessage("Updated Constraint Group, Updating Resource Group with new resources string");
                    OrganizationService.Update(thisGroup);
                    Logger.WriteDebugMessage("Update Resource Group String Successful");
                }

                catch (FaultException<OrganizationServiceFault> ex)
                {
                    Logger.WriteToFile(ex.Message);
                    throw new InvalidPluginExecutionException(ex.Message);
                }
                catch (Exception ex)
                {
                    Logger.WriteToFile(ex.Message);
                    throw new InvalidPluginExecutionException(ex.Message);
                }
            }
        }

        internal void updateTSAfields(Guid tsaGuid)
        {
            CvtHelper.UpdateTSAFields(tsaGuid, Logger, OrganizationService, McsSettings);
        }

        internal void updateMTSAfields(Guid mtsaGuid)
        {
            CvtHelper.UpdateMTSAFields(mtsaGuid, Logger, OrganizationService);
        }
        #endregion


        internal void recurNavigateRules(CalendarRule calRule, Equipment thisEquipment, double effort, int depth)
        {
            depth++;
            if (effort == 0 || calRule.InnerCalendarId == null)
            {
                calRule.IsSimple = true;
                return;
            }
            calRule.IsSimple = null;

            Entity thisInnerCalendarEntity = OrganizationService.Retrieve("calendar", calRule.InnerCalendarId.Id, new ColumnSet(true));
            EntityCollection thisInnerCalendarRules = (EntityCollection)thisInnerCalendarEntity.Attributes["calendarrules"];
            if (thisInnerCalendarRules.Entities.Count == 0)
            {
                Logger.WriteDebugMessage("No Inner Calendar Rules Found");
                return;
            }
            foreach (CalendarRule thisInnerCalRule in thisInnerCalendarRules.Entities)
            {
                if (thisEquipment.cvt_type.Value == 251920000 && thisEquipment.cvt_capacity != null)
                    thisInnerCalRule.Effort = effort;
                else
                    thisInnerCalRule.Effort = 1;

                if (thisInnerCalRule.TimeZoneCode == null)
                    thisInnerCalRule.TimeZoneCode = thisEquipment.TimeZoneCode;
                if (thisInnerCalRule.Rank == null)
                    thisInnerCalRule.Rank = 0;
                if (thisInnerCalRule.Duration == null)
                    thisInnerCalRule.Duration = 540;
                Logger.WriteDebugMessage(String.Format("Rank: {0}, TimeZone: {1}, Duration: {2}", thisInnerCalRule.Rank, thisInnerCalRule.TimeZoneCode, thisInnerCalRule.Duration));
                
                recurNavigateRules(thisInnerCalRule, thisEquipment, effort, depth);
            }
            Logger.WriteDebugMessage("About to Update Inner Calendar " + (depth).ToString());
            OrganizationService.Update(thisInnerCalendarEntity);
            Logger.WriteDebugMessage("updated Inner Calendar");
        }

        internal void ChangeNewlyCreatedCalendarUpdate(Guid sysResourceId)
        {
            Logger.setMethod = "ChangeNewlyCreatedCalendarUpdate";
            Logger.WriteDebugMessage("starting ChangeNewlyCreatedCalendarUpdate");

            using (var srv = new Xrm(OrganizationService))
            {
                try
                {
                    //var defaultDuration = 540;
                    var thisEquipment = srv.EquipmentSet.FirstOrDefault(e => e.Id == sysResourceId);
                    if (thisEquipment == null)
                        return;

                    var resource = srv.mcs_resourceSet.FirstOrDefault(r => r.Id == thisEquipment.mcs_relatedresource.Id);
                    var thisCalendar = OrganizationService.Retrieve("calendar", thisEquipment.CalendarId.Id, new ColumnSet(true));
                    // Retrieve the calendar rules defined in the calendar
                    EntityCollection thiscalendarRules = (EntityCollection)thisCalendar.Attributes["calendarrules"];
                    foreach (CalendarRule calRule in thiscalendarRules.Entities)
                    {
                        if (thisEquipment.cvt_type.Value == 251920000 && thisEquipment.cvt_capacity != null)
                            calRule.Effort = resource.cvt_vistacapacity.Value;

                        if (calRule.InnerCalendarId == null)
                        {
                            calRule.IsSimple = true;
                            break;
                        }
                        else
                        {
                            calRule.IsSimple = null;

                            recurNavigateRules(calRule, thisEquipment, resource.cvt_vistacapacity.Value, 0);
//                            Entity thisInnerCalendarEntity = OrganizationService.Retrieve("calendar", calRule.InnerCalendarId.Id, new ColumnSet(true));
//                            EntityCollection thisInnerCalendarRules = (EntityCollection)thisInnerCalendarEntity.Attributes["calendarrules"];
//                            if (thisInnerCalendarRules.Entities.Count == 0)
//                            {
//                                Logger.WriteDebugMessage("No Inner Calendar Rules Found");
//                                break;
//                            }
//                            foreach (CalendarRule thisInnerCalRule in thisInnerCalendarRules.Entities)
//                            {
////                                CalendarRule thisInnerCalRule = (thisInnerCalendarRules.Entities[0]).ToEntity<CalendarRule>();

//                                if (thisEquipment.cvt_type.Value == 251920000 && thisEquipment.cvt_capacity != null)
//                                    thisInnerCalRule.Effort = resource.cvt_vistacapacity.Value;
//                                else
//                                    thisInnerCalRule.Effort = 1;


//                                if (thisInnerCalRule.TimeZoneCode == null)
//                                    thisInnerCalRule.TimeZoneCode = thisEquipment.TimeZoneCode;
//                                if (thisInnerCalRule.Rank == null)
//                                    thisInnerCalRule.Rank = 0;
//                                if (thisInnerCalRule.Duration == null)
//                                    thisInnerCalRule.Duration = defaultDuration;
//                                Logger.WriteDebugMessage(String.Format("Rank: {0}, TimeZone: {1}, Duration: {2}", thisInnerCalRule.Rank, thisInnerCalRule.TimeZoneCode, thisInnerCalRule.Duration));
                                //}
                        }
                    }
                    Logger.WriteDebugMessage("Updating Equipment Top-Level Calendar");
                    OrganizationService.Update(thisCalendar);
                    Logger.WriteDebugMessage("Calendar Update Complete");
                }
                catch (FaultException<OrganizationServiceFault> ex)
                {
                    Logger.WriteToFile(ex.Message);
                    throw new InvalidPluginExecutionException(ex.Message);
                }
                catch (Exception ex)
                {
                    Logger.WriteToFile(ex.Message);
                    throw new InvalidPluginExecutionException(ex.Message);
                }
            }
        }

        #region Additional Interface Methods
        public override string McsSettingsDebugField
        {
            get { return "mcs_resourceplugin"; }
        }

        public override Entity GetPrimaryEntity()
        {
            return (Entity)PluginExecutionContext.InputParameters["Target"];
        }

        public override Entity GetSecondaryEntity()
        {
            return (Entity)PluginExecutionContext.PostEntityImages["post"];
        }
        #endregion

    }

}