﻿using Microsoft.Xrm.Sdk;
using System;
using System.ServiceModel;
using mcsScheduling;
using System.Linq;
using Microsoft.Xrm.Sdk.Query;
using MCSShared;

namespace MCSPlugins
{
    public class McsSiteCreatePostStageRunner : PluginRunner
    {
        #region Constructor
        public McsSiteCreatePostStageRunner(IServiceProvider serviceProvider)
            : base(serviceProvider)
        {
        }
        #endregion
        #region Internal Methods/Properties
        internal void Execute()
        {
            try
            {
                //use the tracing at the start until the logger is established, 
                //after that the logger contains tracing so just use that.

                TracingService.Trace("at the top");

                if (PluginExecutionContext.InputParameters.Contains("Target") &&
                    PluginExecutionContext.InputParameters["Target"] is Entity)
                {
                    Logger.setMethod = "Execute";
                    Logger.WriteTxnTimingMessage("Starting");

                    Logger.WriteDebugMessage(String.Format("At the top of:{0}", GetType()));
                    using (var srv = new Xrm(OrganizationService))
                    {
                        var thisSite = srv.mcs_siteSet.FirstOrDefault(i => i.Id == PluginExecutionContext.PrimaryEntityId);


                        //Setting attributes for new new System Site. 
                        var systemSite = new Entity("site");
                        systemSite["name"] = McsHelper.getStringValue("mcs_name");
                        systemSite["timezonecode"] = McsHelper.getIntValue("mcs_timezone");

                        //System Site Created. 
                        var newSite = OrganizationService.Create(systemSite);
                        Logger.WriteDebugMessage("New System site created");

                        CvtHelper.AlignLocations(thisSite, OrganizationService, Logger);
                        Logger.WriteDebugMessage("Updated TSS Site's VISN from Facility entered");
                        //Query for the Facilities BU and use that to set the Team's BU
                        var parentFacility = srv.mcs_facilitySet.FirstOrDefault(i => i.Id == thisSite.mcs_FacilityId.Id);

                        //Creating new team based off of the new Site that was created. 
                        var systemTeam = new Entity("team");
                        systemTeam["name"] = McsHelper.getStringValue("mcs_name");
                        if (parentFacility.mcs_BusinessUnitId != null)
                            systemTeam["businessunitid"] = new EntityReference("businessunit", parentFacility.mcs_BusinessUnitId.Id);

                        var newTeam = OrganizationService.Create(systemTeam);
                        Logger.WriteDebugMessage("New Team created");
                        Logger.WriteDebugMessage("About to add Role to Team");
                        // Find the role.
                        QueryExpression query = new QueryExpression
                            {
                                EntityName = Role.EntityLogicalName,
                                ColumnSet = new ColumnSet("roleid"),
                                Criteria = new FilterExpression
                                    {
                                        Conditions =
                                            {

                                                new ConditionExpression
                                                    {
                                                        AttributeName = "name",
                                                        Operator = ConditionOperator.Equal,
                                                        Values = {"VA Site Team"}
                                                    },
                                                new ConditionExpression
                                                    {
                                                        AttributeName = "businessunitid",
                                                        Operator = ConditionOperator.Equal,
                                                        Values = {parentFacility.mcs_BusinessUnitId.Id}
                                                    }
                                            }
                                    }
                            };

                        // Get the role.
                        EntityCollection roles = OrganizationService.RetrieveMultiple(query);
                        if (roles.Entities.Count > 0)
                        {
                            Role salesRole = roles.Entities[0].ToEntity<Role>();
                            var _roleId = salesRole.Id;

                            // Associate the user with the role.
                            if (_roleId != Guid.Empty)
                            {
                                OrganizationService.Associate(
                                    "team",
                                    newTeam,
                                    new Relationship("teamroles_association"),
                                    new EntityReferenceCollection() { new EntityReference(Role.EntityLogicalName, _roleId) });
                            }
                        }
                        Logger.WriteDebugMessage("About to add Role to Team");

                        //Updates the mcs_Site with a reference to the System Site that was created so they are appropiately associated. 
                        Entity mcsSiteUpdate = new Entity("mcs_site");
                        mcsSiteUpdate.Id = PluginExecutionContext.PrimaryEntityId;
                        mcsSiteUpdate["mcs_relatedactualsiteid"] = new EntityReference(systemSite.LogicalName, newSite);
                        OrganizationService.Update(mcsSiteUpdate);
                        Logger.WriteDebugMessage("Mcs Site updated");

                        //Entity teamUpdate = new Entity("team");
                        // teamUpdate.Id = newTeam;
                        // teamUpdate["businessunitid"] = thisSite.mcs_BusinessUnitId.Id;
                        //  OrganizationService.Update(teamUpdate);
                        //  Logger.WriteDebugMessage("Team updated");
                        Logger.WriteTxnTimingMessage("Ending");
                    }
                }
            }
            catch (FaultException<OrganizationServiceFault> ex)
            {
                Logger.WriteToFile(ex.Message);
            }
            catch (Exception ex)
            {
                Logger.WriteToFile(ex.Message);
            }
        }
        #endregion

        #region Additional Interface Methods/Properties
        public override string McsSettingsDebugField
        {
            get { return "mcs_siteplugin"; }
        }

        public override Entity GetPrimaryEntity()
        {
            return (Entity)PluginExecutionContext.InputParameters["Target"];
        }

        public override Entity GetSecondaryEntity()
        {
            return (Entity)PluginExecutionContext.PostEntityImages["post"];
        }
        #endregion
    }
}